import re
from textwrap import dedent
from unittest import TestCase

from pcs.cli import nvset
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)


class NvsetDtoToLines(TestCase):
    def setUp(self):
        self.label = "Meta Attributes"

    def _export(self, dto, with_ids):
        return (
            "\n".join(
                nvset.nvset_dto_to_lines(
                    dto, nvset_label=self.label, with_ids=with_ids
                )
            )
            + "\n"
        )

    def assert_lines(self, dto, lines):
        self.assertEqual(
            self._export(dto, True),
            lines,
        )
        self.assertEqual(
            self._export(dto, False),
            re.sub(r" +\(id:.*\)", "", lines),
        )

    def test_minimal(self):
        dto = CibNvsetDto("my-id", {}, None, [])
        output = dedent(
            f"""\
              {self.label}: my-id
            """
        )
        self.assert_lines(dto, output)

    def test_full(self):
        dto = CibNvsetDto(
            "my-id",
            {"score": "150"},
            CibRuleExpressionDto(
                "my-id-rule",
                CibRuleExpressionType.RULE,
                CibRuleInEffectStatus.UNKNOWN,
                {"boolean-op": "or"},
                None,
                None,
                [
                    CibRuleExpressionDto(
                        "my-id-rule-op",
                        CibRuleExpressionType.OP_EXPRESSION,
                        CibRuleInEffectStatus.UNKNOWN,
                        {"name": "monitor"},
                        None,
                        None,
                        [],
                        "op monitor",
                    ),
                ],
                "op monitor",
            ),
            [
                CibNvpairDto("my-id-pair1", "name1", "value1"),
                CibNvpairDto("my-id-pair2", "name 2", "value 2"),
                CibNvpairDto("my-id-pair3", "name=3", "value=3"),
            ],
        )
        output = dedent(
            f"""\
            {self.label}: my-id score=150
              "name 2"="value 2"
              name1=value1
              "name=3"="value=3"
              Rule: boolean-op=or (id:my-id-rule)
                Expression: op monitor (id:my-id-rule-op)
            """
        )
        self.assert_lines(dto, output)


class NvsetDtoListToLines(TestCase):
    def setUp(self):
        self.label = "Meta Attributes"

    def _export(self, dto, with_ids, include_expired):
        return (
            "\n".join(
                nvset.nvset_dto_list_to_lines(
                    dto,
                    nvset_label=self.label,
                    with_ids=with_ids,
                    include_expired=include_expired,
                )
            )
            + "\n"
        )

    def assert_lines(self, dto, include_expired, lines):
        self.assertEqual(
            self._export(dto, True, include_expired),
            lines,
        )
        self.assertEqual(
            self._export(dto, False, include_expired),
            re.sub(r" +\(id:.*\)", "", lines),
        )

    @staticmethod
    def fixture_dto(in_effect):
        return CibNvsetDto(
            f"id-{in_effect}",
            {"score": "150"},
            CibRuleExpressionDto(
                f"id-{in_effect}-rule",
                CibRuleExpressionType.RULE,
                in_effect,
                {"boolean-op": "or"},
                None,
                None,
                [
                    CibRuleExpressionDto(
                        f"id-{in_effect}-rule-op",
                        CibRuleExpressionType.OP_EXPRESSION,
                        CibRuleInEffectStatus.UNKNOWN,
                        {"name": "monitor"},
                        None,
                        None,
                        [],
                        "op monitor",
                    ),
                ],
                "op monitor",
            ),
            [CibNvpairDto(f"id-{in_effect}-pair1", "name1", "value1")],
        )

    def fixture_dto_list(self):
        return [
            self.fixture_dto(in_effect.value)
            for in_effect in CibRuleInEffectStatus
        ]

    def test_expired_included(self):
        self.maxDiff = None
        output = dedent(
            f"""\
            {self.label} (not yet in effect): id-NOT_YET_IN_EFFECT score=150
              name1=value1
              Rule (not yet in effect): boolean-op=or (id:id-NOT_YET_IN_EFFECT-rule)
                Expression: op monitor (id:id-NOT_YET_IN_EFFECT-rule-op)
            {self.label}: id-IN_EFFECT score=150
              name1=value1
              Rule: boolean-op=or (id:id-IN_EFFECT-rule)
                Expression: op monitor (id:id-IN_EFFECT-rule-op)
            {self.label} (expired): id-EXPIRED score=150
              name1=value1
              Rule (expired): boolean-op=or (id:id-EXPIRED-rule)
                Expression: op monitor (id:id-EXPIRED-rule-op)
            {self.label}: id-UNKNOWN score=150
              name1=value1
              Rule: boolean-op=or (id:id-UNKNOWN-rule)
                Expression: op monitor (id:id-UNKNOWN-rule-op)
        """
        )
        self.assert_lines(self.fixture_dto_list(), True, output)

    def test_expired_excluded(self):
        self.maxDiff = None
        output = dedent(
            f"""\
            {self.label} (not yet in effect): id-NOT_YET_IN_EFFECT score=150
              name1=value1
              Rule (not yet in effect): boolean-op=or (id:id-NOT_YET_IN_EFFECT-rule)
                Expression: op monitor (id:id-NOT_YET_IN_EFFECT-rule-op)
            {self.label}: id-IN_EFFECT score=150
              name1=value1
              Rule: boolean-op=or (id:id-IN_EFFECT-rule)
                Expression: op monitor (id:id-IN_EFFECT-rule-op)
            {self.label}: id-UNKNOWN score=150
              name1=value1
              Rule: boolean-op=or (id:id-UNKNOWN-rule)
                Expression: op monitor (id:id-UNKNOWN-rule-op)
        """
        )
        self.assert_lines(self.fixture_dto_list(), False, output)



# pylint: disable=protected-access
import dataclasses
from queue import Empty
from unittest import mock

from tornado.testing import gen_test

from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
    TaskResultDto,
)
from pcs.common.async_tasks.types import (
    TaskFinishType,
    TaskKillReason,
    TaskState,
)
from pcs.common.reports import ReportItem
from pcs.common.reports.messages import CibUpgradeSuccessful
from pcs.daemon.async_tasks import scheduler
from pcs.daemon.async_tasks.task import (
    Task,
    TaskConfig,
)
from pcs.daemon.async_tasks.worker.executor import task_executor
from pcs.daemon.async_tasks.worker.types import (
    Message,
    TaskExecuted,
)

from .helpers import (
    ANOTHER_AUTH_USER,
    AUTH_USER,
    SchedulerBaseAsyncTestCase,
    SchedulerBaseTestCase,
)

WORKER1_PID = 2222
WORKER2_PID = 3333


class GetTaskTest(SchedulerBaseTestCase):
    def test_task_exists(self):
        self._create_tasks(1)
        task_result = self.scheduler.get_task("id0", AUTH_USER)
        self.assertEqual("command 0", task_result.command.command_name)

    def test_task_not_exists(self):
        self._create_tasks(1)
        with self.assertRaises(scheduler.TaskNotFoundError):
            self.scheduler.get_task("nonexistent", AUTH_USER)

    def test_delete_finished(self):
        self._create_tasks(3)
        self.scheduler._task_register["id1"].state = TaskState.FINISHED
        self.assertIsNone(
            self.scheduler._task_register["id1"]._to_delete_timestamp
        )
        self.scheduler.get_task("id1", AUTH_USER)
        self.assertIsNotNone(
            self.scheduler._task_register["id1"]._to_delete_timestamp
        )

    def test_different_user(self):
        self._create_tasks(1)
        with self.assertRaises(scheduler.TaskNotFoundError):
            self.scheduler.get_task("id0", ANOTHER_AUTH_USER)


class KillTaskTest(SchedulerBaseTestCase):
    def test_task_exists(self):
        self._create_tasks(2)
        self.scheduler.kill_task("id0", AUTH_USER)
        task1_dto = self.scheduler.get_task("id0", AUTH_USER)
        task2_dto = self.scheduler.get_task("id1", AUTH_USER)
        # State and task_finish_type is not changing, only kill_reason
        self.assertEqual(TaskKillReason.USER, task1_dto.kill_reason)
        self.assertEqual(TaskState.CREATED, task1_dto.state)
        self.assertEqual(TaskFinishType.UNFINISHED, task1_dto.task_finish_type)
        self.assertIsNone(task2_dto.kill_reason)
        self.assertEqual(TaskState.CREATED, task2_dto.state)
        self.assertEqual(TaskFinishType.UNFINISHED, task2_dto.task_finish_type)

    def test_task_not_exists(self):
        self._create_tasks(1)
        with self.assertRaises(scheduler.TaskNotFoundError):
            self.scheduler.kill_task("nonexistent", AUTH_USER)

    def test_different_user(self):
        self._create_tasks(1)
        with self.assertRaises(scheduler.TaskNotFoundError):
            self.scheduler.kill_task("id0", ANOTHER_AUTH_USER)


class NewTaskTest(SchedulerBaseTestCase):
    def test_new_task(self):
        self._create_tasks(1)
        self.assertEqual(
            TaskResultDto(
                "id0",
                CommandDto(
                    "command 0", {}, CommandOptionsDto(request_timeout=None)
                ),
                [],
                TaskState.CREATED,
                TaskFinishType.UNFINISHED,
                None,
                None,
            ),
            self.scheduler.get_task("id0", AUTH_USER),
        )
        self.assertListEqual(
            ["id0"],
            list(self.scheduler._task_register.keys()),
        )

    class UuidStub:
        def __init__(self, uuid_list):
            self._iter = iter(uuid_list)

        @property
        def hex(self):
            return next(self._iter)

    @mock.patch("uuid.uuid4", new=UuidStub(["id0", "id0", "id1"]))
    def test_duplicate_task_ident(self):
        self._create_tasks(2)
        self.assertListEqual(
            ["id0", "id1"],
            list(self.scheduler._task_register.keys()),
        )


class ReceiveMessagesTest(SchedulerBaseAsyncTestCase):
    @gen_test
    async def test_wrong_payload_type(self):
        worker_com = self.worker_com
        self._create_tasks(2)
        worker_com.put(Message("id0", TaskExecuted(WORKER1_PID)))
        worker_com.put(Message("id0", "definitely not a payload"))
        received = 0
        while received < 2:
            received += await self.scheduler._receive_messages()
        task1_dto = self.scheduler.get_task("id0", AUTH_USER)
        task2_dto = self.scheduler.get_task("id1", AUTH_USER)
        self.assertEqual(
            TaskKillReason.INTERNAL_MESSAGING_ERROR, task1_dto.kill_reason
        )
        self.assertIsNone(task2_dto.kill_reason)

    @gen_test
    async def test_wrong_message_id(self):
        # TODO
        pass

    @gen_test
    async def test_wrong_message_type(self):
        self._create_tasks(2)
        self.worker_com.put(Message("id0", TaskExecuted(WORKER1_PID)))
        self.worker_com.put("definitely not a message")
        received = 0
        while received < 2:
            received += await self.scheduler._receive_messages()
        self.logger_mock.error.assert_called_once()

    @gen_test
    async def test_all_messages_consumed(self):
        task_count = 3
        message_count = 3
        self._create_tasks(task_count)
        for _ in range(message_count):
            for idx in range(task_count):
                self.worker_com.put(
                    Message(
                        f"id{idx}",
                        ReportItem.error(CibUpgradeSuccessful()).to_dto(),
                    )
                )
        received = 0
        while received < task_count * message_count:
            received += await self.scheduler._receive_messages()
        with self.assertRaises(Empty):
            self.worker_com.get_nowait()


class ProcessTasksTest(SchedulerBaseAsyncTestCase):
    @gen_test
    async def test_empty_created_task_index(self):
        await self.scheduler._process_tasks()
        self.mp_pool_mock.assert_not_called()

    @gen_test
    async def test_normal_run(self):
        task_count = 4
        self._create_tasks(task_count)
        await self.scheduler._process_tasks()
        for task_ident in [f"id{i}" for i in range(task_count)]:
            self.assertEqual(
                TaskState.QUEUED,
                self.scheduler.get_task(task_ident, AUTH_USER).state,
            )
        self.assertEqual(
            0,
            len(
                list(
                    task
                    for task in self.scheduler._task_register.values()
                    if task.state == TaskState.CREATED
                )
            ),
        )

    @gen_test
    async def test_killed_task(self):
        self._create_tasks(3)
        tasks = list(self.scheduler._task_register.values())
        self.scheduler.kill_task("id1", AUTH_USER)
        await self.scheduler._process_tasks()
        deleted_task = self.scheduler.get_task("id1", AUTH_USER)
        self.assertEqual(TaskFinishType.KILL, deleted_task.task_finish_type)
        self.assertEqual(TaskState.FINISHED, deleted_task.state)
        self.assertEqual(
            TaskState.QUEUED,
            self.scheduler.get_task("id0", AUTH_USER).state,
        )
        self.assertEqual(
            TaskState.QUEUED,
            self.scheduler.get_task("id2", AUTH_USER).state,
        )
        self.mp_pool_mock.apply_async.assert_has_calls(
            [
                mock.call(
                    func=task_executor,
                    args=[tasks[0].to_worker_command()],
                ),
                mock.call(
                    func=task_executor,
                    args=[tasks[2].to_worker_command()],
                ),
            ]
        )


def get_generator(return_values):
    def generator():
        for item in return_values:
            yield item

    gen = generator()
    return lambda *args, **kwargs: next(gen)


class DeadlockDetectionTest(SchedulerBaseTestCase):
    @staticmethod
    def _create_task(index, state=TaskState.CREATED):
        task = Task(
            f"id{index}",
            CommandDto(f"cmd{index}", {}, CommandOptionsDto()),
            AUTH_USER,
            TaskConfig(),
        )
        if state != TaskState.CREATED:
            task.state = state
        return task

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: False)
    def test_threshold_not_achieved(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2")
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2)
        }
        self.assertFalse(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", get_generator([True, False]))
    def test_threshold_not_achieved_for_some(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.EXECUTED)
        task3 = self._create_task("3")
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2, task3)
        }
        self.assertFalse(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_no_tasks_waiting(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.EXECUTED)
        task3 = self._create_task("3", TaskState.FINISHED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2, task3)
        }
        self.assertFalse(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_new_tasks_waiting(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.CREATED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2)
        }
        self.assertTrue(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_queued_tasks_waiting(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.QUEUED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2)
        }
        self.assertTrue(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_any_tasks_waiting(self):
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.QUEUED)
        task3 = self._create_task("3", TaskState.CREATED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2, task3)
        }
        self.assertTrue(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_workers_available(self):
        self.scheduler._config = dataclasses.replace(
            self.scheduler._config, worker_count=3
        )
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.QUEUED)
        task3 = self._create_task("3", TaskState.CREATED)
        task4 = self._create_task("4", TaskState.EXECUTED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2, task3, task4)
        }
        self.assertFalse(self.scheduler._is_possibly_dead_locked())

    @mock.patch.object(Task, "is_defunct", lambda self, timeout: True)
    def test_workers_available_with_tmp_workers(self):
        self.scheduler._config = dataclasses.replace(
            self.scheduler._config, worker_count=1
        )
        self.scheduler._single_use_process_pool = list(range(2))
        task1 = self._create_task("1", TaskState.EXECUTED)
        task2 = self._create_task("2", TaskState.QUEUED)
        task3 = self._create_task("3", TaskState.CREATED)
        task4 = self._create_task("4", TaskState.EXECUTED)
        self.scheduler._task_register = {
            task.task_ident: task for task in (task1, task2, task3, task4)
        }
        self.assertFalse(self.scheduler._is_possibly_dead_locked())



from typing import (
    Any,
    List,
)

from pcs import resource
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.common.routing import (
    CliCmdInterface,
    create_router,
)


def resource_defaults_cmd(parent_cmd: List[str]) -> CliCmdInterface:
    def _get_router(
        lib: Any, argv: List[str], modifiers: InputModifiers
    ) -> None:
        """
        Options:
          * -f - CIB file
          * --force - allow unknown options
        """
        if argv and "=" in argv[0]:
            # DEPRECATED legacy command
            return resource.resource_defaults_legacy_cmd(
                lib, argv, modifiers, deprecated_syntax_used=True
            )

        router = create_router(
            {
                "config": resource.resource_defaults_config_cmd,
                "set": create_router(
                    {
                        "create": resource.resource_defaults_set_create_cmd,
                        "delete": resource.resource_defaults_set_remove_cmd,
                        "remove": resource.resource_defaults_set_remove_cmd,
                        "update": resource.resource_defaults_set_update_cmd,
                    },
                    parent_cmd + ["set"],
                ),
                "update": resource.resource_defaults_legacy_cmd,
            },
            parent_cmd,
            default_cmd="config",
        )
        return router(lib, argv, modifiers)

    return _get_router


def resource_op_defaults_cmd(parent_cmd: List[str]) -> CliCmdInterface:
    def _get_router(
        lib: Any, argv: List[str], modifiers: InputModifiers
    ) -> None:
        """
        Options:
          * -f - CIB file
          * --force - allow unknown options
        """
        if argv and "=" in argv[0]:
            # DEPRECATED legacy command
            return resource.resource_op_defaults_legacy_cmd(
                lib, argv, modifiers, deprecated_syntax_used=True
            )

        router = create_router(
            {
                "config": resource.resource_op_defaults_config_cmd,
                "set": create_router(
                    {
                        "create": resource.resource_op_defaults_set_create_cmd,
                        "delete": resource.resource_op_defaults_set_remove_cmd,
                        "remove": resource.resource_op_defaults_set_remove_cmd,
                        "update": resource.resource_op_defaults_set_update_cmd,
                    },
                    parent_cmd + ["set"],
                ),
                "update": resource.resource_op_defaults_legacy_cmd,
            },
            parent_cmd,
            default_cmd="config",
        )
        return router(lib, argv, modifiers)

    return _get_router



from typing import NewType

from pcs.common.tools import Version

PcmkRoleType = NewType("PcmkRoleType", str)
PcmkOnFailAction = NewType("PcmkOnFailAction", str)

PCMK_ROLE_STARTED = PcmkRoleType("Started")
PCMK_ROLE_STOPPED = PcmkRoleType("Stopped")
PCMK_ROLE_PROMOTED = PcmkRoleType("Promoted")
PCMK_ROLE_UNPROMOTED = PcmkRoleType("Unpromoted")
PCMK_ROLE_PROMOTED_LEGACY = PcmkRoleType("Master")
PCMK_ROLE_UNPROMOTED_LEGACY = PcmkRoleType("Slave")
PCMK_ON_FAIL_ACTION_IGNORE = PcmkOnFailAction("ignore")
PCMK_ON_FAIL_ACTION_BLOCK = PcmkOnFailAction("block")
PCMK_ON_FAIL_ACTION_DEMOTE = PcmkOnFailAction("demote")
PCMK_ON_FAIL_ACTION_STOP = PcmkOnFailAction("stop")
PCMK_ON_FAIL_ACTION_RESTART = PcmkOnFailAction("restart")
PCMK_ON_FAIL_ACTION_STANDBY = PcmkOnFailAction("standby")
PCMK_ON_FAIL_ACTION_FENCE = PcmkOnFailAction("fence")
PCMK_ON_FAIL_ACTION_RESTART_CONTAINER = PcmkOnFailAction("restart-container")
PCMK_ROLE_PROMOTED_PRIMARY = PCMK_ROLE_PROMOTED
PCMK_ROLE_UNPROMOTED_PRIMARY = PCMK_ROLE_UNPROMOTED
PCMK_ROLES_PROMOTED = (PCMK_ROLE_PROMOTED, PCMK_ROLE_PROMOTED_LEGACY)
PCMK_ROLES_UNPROMOTED = (PCMK_ROLE_UNPROMOTED, PCMK_ROLE_UNPROMOTED_LEGACY)
PCMK_ROLES_RUNNING = (
    (PCMK_ROLE_STARTED,) + PCMK_ROLES_PROMOTED + PCMK_ROLES_UNPROMOTED
)
PCMK_ROLES = (PCMK_ROLE_STOPPED,) + PCMK_ROLES_RUNNING
PCMK_ACTION_START = "start"
PCMK_ACTIONS = (PCMK_ACTION_START, "stop", "promote", "demote")
PCMK_NEW_ROLES_CIB_VERSION = Version(3, 7, 0)
PCMK_RULES_NODE_ATTR_EXPR_WITH_INT_TYPE_CIB_VERSION = Version(3, 5, 0)
PCMK_ON_FAIL_DEMOTE_CIB_VERSION = Version(3, 4, 0)



from typing import cast

from lxml.etree import _Element

from pcs.common.pacemaker.rule import (
    CibRuleDateCommonDto,
    CibRuleExpressionDto,
)
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)
from pcs.lib.xml_tools import export_attributes

from .cib_to_str import RuleToStr
from .in_effect import RuleInEffectEval


def rule_element_to_dto(
    in_effect_eval: RuleInEffectEval, rule_el: _Element
) -> CibRuleExpressionDto:
    """
    Export a rule xml element including its children to their DTOs

    in_effect_eval -- a class for evaluating if a rule is in effect
    rule_el -- the rule to be converted to DTO
    """
    return _Exporter(in_effect_eval, RuleToStr()).export(rule_el)


class _Exporter:
    _tag_to_type = {
        "rule": CibRuleExpressionType.RULE,
        "expression": CibRuleExpressionType.EXPRESSION,
        "date_expression": CibRuleExpressionType.DATE_EXPRESSION,
        "op_expression": CibRuleExpressionType.OP_EXPRESSION,
        "rsc_expression": CibRuleExpressionType.RSC_EXPRESSION,
    }

    def __init__(self, in_effect_eval: RuleInEffectEval, str_eval: RuleToStr):
        self._in_effect_eval = in_effect_eval
        self._str_eval = str_eval

    def export(self, rule_el: _Element) -> CibRuleExpressionDto:
        return self._tag_to_export[str(rule_el.tag)](self, rule_el)

    def _rule_to_dto(self, rule_el: _Element) -> CibRuleExpressionDto:
        children_dto_list = [
            self._tag_to_export[str(child.tag)](self, child)
            # The xpath method has a complicated return value, but we know our
            # xpath expression only returns elements.
            for child in cast(_Element, rule_el.xpath(self._xpath_for_export))
        ]
        rule_id = str(rule_el.get("id", ""))
        return CibRuleExpressionDto(
            rule_id,
            self._tag_to_type[str(rule_el.tag)],
            self._in_effect_eval.get_rule_status(rule_id),
            export_attributes(rule_el, with_id=False),
            None,
            None,
            children_dto_list,
            self._str_eval.get_str(rule_el),
        )

    def _common_expr_to_dto(self, expr_el: _Element) -> CibRuleExpressionDto:
        return CibRuleExpressionDto(
            str(expr_el.get("id", "")),
            self._tag_to_type[str(expr_el.tag)],
            CibRuleInEffectStatus.UNKNOWN,
            export_attributes(expr_el, with_id=False),
            None,
            None,
            [],
            self._str_eval.get_str(expr_el),
        )

    @staticmethod
    def _date_common_to_dto(expr_el: _Element) -> CibRuleDateCommonDto:
        return CibRuleDateCommonDto(
            str(expr_el.get("id", "")),
            export_attributes(expr_el, with_id=False),
        )

    def _date_expr_to_dto(self, expr_el: _Element) -> CibRuleExpressionDto:
        date_spec = expr_el.find("./date_spec")
        duration = expr_el.find("./duration")
        return CibRuleExpressionDto(
            str(expr_el.get("id", "")),
            self._tag_to_type[str(expr_el.tag)],
            CibRuleInEffectStatus.UNKNOWN,
            export_attributes(expr_el, with_id=False),
            None if date_spec is None else self._date_common_to_dto(date_spec),
            None if duration is None else self._date_common_to_dto(duration),
            [],
            self._str_eval.get_str(expr_el),
        )

    _tag_to_export = {
        "rule": _rule_to_dto,
        "expression": _common_expr_to_dto,
        "date_expression": _date_expr_to_dto,
        "op_expression": _common_expr_to_dto,
        "rsc_expression": _common_expr_to_dto,
    }

    _xpath_for_export = "./*[{export_tags}]".format(
        export_tags=" or ".join(f"self::{tag}" for tag in _tag_to_export)
    )



from typing import Sequence


class PrintableTreeNode:
    @property
    def members(self) -> Sequence["PrintableTreeNode"]:
        raise NotImplementedError()

    @property
    def detail(self) -> list[str]:
        raise NotImplementedError()

    @property
    def is_leaf(self) -> bool:
        raise NotImplementedError()

    def get_title(self, verbose: bool) -> str:
        raise NotImplementedError()


def tree_to_lines(
    node: PrintableTreeNode,
    verbose: bool = False,
    title_prefix: str = "",
    indent: str = "",
) -> list[str]:
    """
    Return sequence of strings representing lines to print out tree structure on
    command line.
    """
    result = []
    note = ""
    if node.is_leaf:
        note = " [displayed elsewhere]"
    title = node.get_title(verbose)
    result.append(f"{title_prefix}{title}{note}")
    if node.is_leaf:
        return result
    _indent = "|  "
    if not node.members:
        _indent = "   "
    for line in node.detail:
        result.append(f"{indent}{_indent}{line}")
    _indent = "|  "
    _title_prefix = "|- "
    for member in node.members:
        if member == node.members[-1]:
            _indent = "   "
            _title_prefix = "`- "
        result.extend(
            tree_to_lines(
                member,
                verbose,
                indent=f"{indent}{_indent}",
                title_prefix=f"{indent}{_title_prefix}",
            )
        )
    return result



from typing import (
    List,
    NamedTuple,
)

from pcs.common.dr import DrRole
from pcs.common.types import StringIterable
from pcs.lib.interface.config import FacadeInterface


class DrSite(NamedTuple):
    role: DrRole
    node_name_list: List[str]


class Facade(FacadeInterface):
    @classmethod
    def create(cls, local_role: DrRole) -> "Facade":
        return cls(
            dict(
                local=dict(
                    role=local_role.value,
                ),
                remote_sites=[],
            )
        )

    @classmethod
    def empty(cls) -> "Facade":
        return cls({})

    @property
    def local_role(self) -> DrRole:
        return DrRole(self._config["local"]["role"])

    def add_site(self, role: DrRole, node_list: StringIterable) -> None:
        self._config["remote_sites"].append(
            dict(
                role=role.value,
                nodes=[dict(name=node) for node in node_list],
            )
        )

    def get_remote_site_list(self) -> List[DrSite]:
        return [
            DrSite(
                DrRole(conf_site["role"]),
                [node["name"] for node in conf_site["nodes"]],
            )
            for conf_site in self._config.get("remote_sites", [])
        ]



from unittest import TestCase

from pcs.common import reports
from pcs.lib.commands import tag as cmd_tag

from pcs_test.tier0.lib.commands.tag.tag_common import (
    fixture_resources_for_reference_ids,
    fixture_tags_xml,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


# This class does not focusing on validation testing, there are validator tests
# for that in pcs_test.tier0.lib.cib.test_tag
class TestTagUpdate(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            resources=fixture_resources_for_reference_ids(
                ["e1", "e2", "e3", "a", "b"]
            ),
            tags=fixture_tags_xml([("t", ["e1", "e2", "e3"])]),
        )

    def test_add_ids(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e1", "e2", "e3", "a", "b"])]),
        )
        cmd_tag.update(self.env_assist.get_env(), "t", ["a", "b"], [])

    def test_add_ids_before(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["a", "b", "e1", "e2", "e3"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["a", "b"],
            [],
            adjacent_idref="e1",
        )

    def test_add_ids_after(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e1", "b", "a", "e2", "e3"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["b", "a"],
            [],
            adjacent_idref="e1",
            put_after_adjacent=True,
        )

    def test_remove_ids(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2"])]),
        )
        cmd_tag.update(self.env_assist.get_env(), "t", [], ["e1", "e3"])

    def test_combination_add_remove(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2", "a", "b"])]),
        )
        cmd_tag.update(self.env_assist.get_env(), "t", ["a", "b"], ["e1", "e3"])

    def test_combination_add_before_remove(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["a", "b", "e2"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["a", "b"],
            ["e1", "e3"],
            adjacent_idref="e2",
        )

    def test_combination_add_after_remove(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2", "a", "b"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["a", "b"],
            ["e1", "e3"],
            adjacent_idref="e2",
            put_after_adjacent=True,
        )

    def test_move_existing_before(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2", "e3", "e1"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["e2", "e3"],
            [],
            adjacent_idref="e1",
        )

    def test_move_existing_after(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2", "e3", "e1"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["e1"],
            [],
            adjacent_idref="e3",
            put_after_adjacent=True,
        )

    def test_move_new_and_existing_before(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e2", "a", "e3", "b", "e1"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["e2", "a", "e3", "b"],
            [],
            adjacent_idref="e1",
        )

    def test_move_new_and_existing_after(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e1", "e2", "b", "a", "e3"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["b", "a", "e3"],
            [],
            adjacent_idref="e2",
            put_after_adjacent=True,
        )

    def test_move_new_and_existing_before_and_remove(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["a", "b", "e2", "e3"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["a", "b"],
            ["e1"],
            adjacent_idref="e2",
        )

    def test_move_new_and_existing_after_and_remove(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["e1", "e2", "b", "a"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["b", "a"],
            ["e3"],
            adjacent_idref="e2",
            put_after_adjacent=True,
        )

    def test_remove_all_existing_but_add_new_ones(self):
        self.config.env.push_cib(
            tags=fixture_tags_xml([("t", ["a", "b"])]),
        )
        cmd_tag.update(
            self.env_assist.get_env(),
            "t",
            ["a", "b"],
            ["e1", "e2", "e3"],
        )
        self.env_assist.assert_reports([])

    def test_raises_exception_in_case_of_report(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.update(
                self.env_assist.get_env(),
                "t",
                [],
                ["e1", "e2", "e3"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_REFERENCES_WITHOUT_REMOVING_TAG,
                    tag_id="t",
                )
            ]
        )



from unittest import (
    TestCase,
    mock,
)

from pcs import quorum
from pcs.cli.common.errors import CmdLineInputError

from pcs_test.tools.misc import dict_to_modifiers


class LocalQnetdCertsMixin:
    def _call_cmd(self, argv, modifiers=None):
        self.cli_command(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertEqual(
            cm.exception.message,
            "Expected arguments: <qnetd_host> <cluster_name>",
        )
        self.lib_command.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a", "b", "c"])
        self.assertEqual(
            cm.exception.message,
            "Expected arguments: <qnetd_host> <cluster_name>",
        )
        self.lib_command.assert_not_called()

    def test_empty_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["", ""])
        self.assertEqual(
            cm.exception.message,
            "Expected arguments: <qnetd_host> <cluster_name>",
        )
        self.lib_command.assert_not_called()


class CheckLocalQnetdCerts(LocalQnetdCertsMixin, TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["quorum"])
        self.lib.quorum = mock.Mock()
        self.lib_command = self.lib.quorum.device_net_certificate_check_local
        self.cli_command = quorum.check_local_qnetd_certs_cmd

    @mock.patch("pcs.quorum.print")
    def test_success_true(self, mock_print):
        self.lib_command.return_value = True
        self._call_cmd(["a", "b"])
        self.lib_command.assert_called_once_with("a", "b")
        mock_print.assert_called_once_with("certificate present")

    @mock.patch("pcs.quorum.print")
    def test_success_false(self, mock_print):
        self.lib_command.return_value = False
        self._call_cmd(["a", "b"])
        self.lib_command.assert_called_once_with("a", "b")
        mock_print.assert_called_once_with("certificate missing")


class SetupLocalQnetdCerts(LocalQnetdCertsMixin, TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["quorum"])
        self.lib.quorum = mock.Mock()
        self.lib_command = self.lib.quorum.device_net_certificate_setup_local
        self.cli_command = quorum.setup_local_qnetd_certs_cmd

    def test_success(self):
        self._call_cmd(["a", "b"])
        self.lib_command.assert_called_once_with("a", "b")



from dataclasses import dataclass
from typing import (
    Mapping,
    Optional,
    Sequence,
)

from pcs.common.interface.dto import DataTransferObject
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)


@dataclass(frozen=True)
class CibRuleDateCommonDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    options: Mapping[str, str]


@dataclass(frozen=True)
class CibRuleExpressionDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    id: str  # pylint: disable=invalid-name
    type: CibRuleExpressionType
    in_effect: CibRuleInEffectStatus  # only valid for type==rule
    options: Mapping[str, str]
    date_spec: Optional[CibRuleDateCommonDto]
    duration: Optional[CibRuleDateCommonDto]
    expressions: Sequence["CibRuleExpressionDto"]
    as_string: str



from unittest import TestCase

from pcs.common.dr import DrRole
from pcs.lib.dr.config import facade


class Facade(TestCase):
    def test_create(self):
        for role in DrRole:
            with self.subTest(local_role=role.value):
                self.assertEqual(
                    dict(
                        local=dict(
                            role=role.value,
                        ),
                        remote_sites=[],
                    ),
                    facade.Facade.create(role).config,
                )

    def test_local_role(self):
        for role in DrRole:
            with self.subTest(local_role=role.value):
                cfg = facade.Facade(
                    {
                        "local": {
                            "role": role.value,
                        },
                        "remote_sites": [],
                    }
                )
                self.assertEqual(cfg.local_role, role)

    def test_add_site(self):
        node_list = [f"node{i}" for i in range(4)]
        cfg = facade.Facade.create(DrRole.PRIMARY)
        cfg.add_site(DrRole.RECOVERY, node_list)
        self.assertEqual(
            dict(
                local=dict(
                    role=DrRole.PRIMARY.value,
                ),
                remote_sites=[
                    dict(
                        role=DrRole.RECOVERY.value,
                        nodes=[dict(name=node) for node in node_list],
                    ),
                ],
            ),
            cfg.config,
        )


class GetRemoteSiteList(TestCase):
    def test_no_sites(self):
        cfg = facade.Facade(
            {
                "local": {
                    "role": DrRole.PRIMARY.value,
                },
                "remote_sites": [],
            }
        )
        self.assertEqual(cfg.get_remote_site_list(), [])

    def test_one_site(self):
        cfg = facade.Facade(
            {
                "local": {
                    "role": DrRole.PRIMARY.value,
                },
                "remote_sites": [
                    {
                        "role": DrRole.RECOVERY.value,
                        "nodes": [
                            {"name": "node1"},
                        ],
                    },
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [
                facade.DrSite(role=DrRole.RECOVERY, node_name_list=["node1"]),
            ],
        )

    def test_more_sites(self):
        cfg = facade.Facade(
            {
                "local": {
                    "role": DrRole.RECOVERY.value,
                },
                "remote_sites": [
                    {
                        "role": DrRole.PRIMARY.value,
                        "nodes": [
                            {"name": "nodeA1"},
                            {"name": "nodeA2"},
                        ],
                    },
                    {
                        "role": DrRole.RECOVERY.value,
                        "nodes": [
                            {"name": "nodeB1"},
                            {"name": "nodeB2"},
                        ],
                    },
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [
                facade.DrSite(
                    role=DrRole.PRIMARY, node_name_list=["nodeA1", "nodeA2"]
                ),
                facade.DrSite(
                    role=DrRole.RECOVERY, node_name_list=["nodeB1", "nodeB2"]
                ),
            ],
        )

    def test_no_nodes(self):
        cfg = facade.Facade(
            {
                "local": {
                    "role": DrRole.PRIMARY.value,
                },
                "remote_sites": [
                    {
                        "role": DrRole.RECOVERY.value,
                        "nodes": [],
                    },
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [
                facade.DrSite(role=DrRole.RECOVERY, node_name_list=[]),
            ],
        )



from typing import (
    Mapping,
    Optional,
    Type,
)

from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.lib.corosync import constants
from pcs.lib.interface.config import (
    ExporterInterface,
    ParserErrorException,
    ParserInterface,
)

AttrName = str
AttrValue = str
AttrDict = dict[AttrName, AttrValue]
AttrTuple = tuple[AttrName, AttrValue]


class Section:
    def __init__(self, name: str):
        self._parent: Optional["Section"] = None
        self._attr_list: list[AttrTuple] = []
        self._section_list: list["Section"] = []
        self._name: str = name

    @property
    def parent(self) -> Optional["Section"]:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def empty(self) -> bool:
        return not self._attr_list and not self._section_list

    def export(self, indent: str = "    ") -> str:
        lines = []
        for attr in self._attr_list:
            lines.append("{0}: {1}".format(*attr))
        if self._attr_list and self._section_list:
            lines.append("")
        section_count = len(self._section_list)
        for index, section in enumerate(self._section_list, 1):
            lines.extend(str(section).split("\n"))
            if not lines[-1].strip():
                del lines[-1]
            if index < section_count:
                lines.append("")
        if self.parent:
            lines = [indent + x if x else x for x in lines]
            lines.insert(0, self.name + " {")
            lines.append("}")
        final = "\n".join(lines)
        if final:
            final += "\n"
        return final

    def get_root(self) -> "Section":
        parent = self
        while parent.parent:
            parent = parent.parent
        return parent

    def get_attributes(
        self, name: Optional[AttrName] = None
    ) -> list[AttrTuple]:
        return [
            attr for attr in self._attr_list if name is None or attr[0] == name
        ]

    def get_attributes_dict(self) -> AttrDict:
        return {attr[0]: attr[1] for attr in self._attr_list}

    def get_attribute_value(
        self, name: AttrName, default: Optional[AttrValue] = None
    ) -> Optional[AttrValue]:
        return self.get_attributes_dict().get(name, default)

    def add_attribute(self, name: AttrName, value: AttrValue) -> "Section":
        self._attr_list.append((name, value))
        return self

    def del_attributes_by_name(
        self, name: AttrName, value: Optional[AttrValue] = None
    ) -> "Section":
        self._attr_list = [
            attr
            for attr in self._attr_list
            if not (attr[0] == name and (value is None or attr[1] == value))
        ]
        return self

    def set_attribute(self, name: AttrName, value: AttrValue) -> "Section":
        found = False
        new_attr_list = []
        for attr in self._attr_list:
            if attr[0] != name:
                new_attr_list.append(attr)
            elif not found:
                found = True
                new_attr_list.append((name, value))
        self._attr_list = new_attr_list
        if not found:
            self.add_attribute(name, value)
        return self

    def get_sections(self, name: Optional[str] = None) -> list["Section"]:
        return [
            section
            for section in self._section_list
            if name is None or section.name == name
        ]

    def add_section(self, section: "Section") -> "Section":
        parent: Optional["Section"] = self
        while parent:
            if parent == section:
                raise CircularParentshipException()
            parent = parent.parent
        if section.parent:
            section.parent.del_section(section)
        # here we are editing obj's _parent attribute of the same class
        # pylint: disable=protected-access
        section._parent = self
        self._section_list.append(section)
        return self

    def del_section(self, section: "Section") -> "Section":
        self._section_list.remove(section)
        # don't set parent to None if the section was not found in the list
        # thanks to remove raising a ValueError in that case
        # here we are editing obj's _parent attribute of the same class
        # pylint: disable=protected-access
        section._parent = None
        return self

    def __str__(self) -> str:
        return self.export()


class Parser(ParserInterface):
    @staticmethod
    def parse(raw_file_data: bytes) -> Section:
        root = Section("")
        Parser._parse_section(raw_file_data.decode("utf-8").split("\n"), root)
        return root

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: file_type_codes.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        # TODO switch to new exceptions / reports and do not ignore input
        # arguments of the function
        return [
            reports.ReportItem.error(
                Parser.parser_exception_to_report_msg(exception)
            )
        ]

    @staticmethod
    def _parse_section(lines: list[str], section: Section) -> None:
        # parser should work the same way as the original parser in corosync
        while lines:
            current_line = lines.pop(0).strip()
            if not current_line or current_line[0] == "#":
                continue
            if "{" in current_line:
                section_name_candidate, after_brace_junk = current_line.rsplit(
                    "{", 1
                )
                if after_brace_junk.strip():
                    raise ExtraCharactersAfterOpeningBraceException()
                section_name = section_name_candidate.strip()
                if not section_name:
                    raise MissingSectionNameBeforeOpeningBraceException()
                new_section = Section(section_name.strip())
                section.add_section(new_section)
                Parser._parse_section(lines, new_section)
            elif "}" in current_line:
                if current_line.strip() != "}":
                    raise ExtraCharactersBeforeOrAfterClosingBraceException()
                if not section.parent:
                    raise UnexpectedClosingBraceException()
                return
            elif ":" in current_line:
                section.add_attribute(
                    *[x.strip() for x in current_line.split(":", 1)]
                )
            else:
                raise LineIsNotSectionNorKeyValueException()
        if section.parent:
            raise MissingClosingBraceException()

    @staticmethod
    def parser_exception_to_report_msg(
        exception: ParserErrorException,
    ) -> reports.item.ReportItemMessage:
        exc_to_msg: Mapping[
            Type[ParserErrorException], Type[reports.item.ReportItemMessage]
        ] = {
            MissingClosingBraceException: (
                reports.messages.ParseErrorCorosyncConfMissingClosingBrace
            ),
            UnexpectedClosingBraceException: (
                reports.messages.ParseErrorCorosyncConfUnexpectedClosingBrace
            ),
            MissingSectionNameBeforeOpeningBraceException: (
                reports.messages.ParseErrorCorosyncConfMissingSectionNameBeforeOpeningBrace
            ),
            ExtraCharactersAfterOpeningBraceException: (
                reports.messages.ParseErrorCorosyncConfExtraCharactersAfterOpeningBrace
            ),
            ExtraCharactersBeforeOrAfterClosingBraceException: (
                reports.messages.ParseErrorCorosyncConfExtraCharactersBeforeOrAfterClosingBrace
            ),
            LineIsNotSectionNorKeyValueException: (
                reports.messages.ParseErrorCorosyncConfLineIsNotSectionNorKeyValue
            ),
        }
        return exc_to_msg.get(
            type(exception), reports.messages.ParseErrorCorosyncConf
        )()


class Exporter(ExporterInterface):
    @staticmethod
    def export(config_structure: Section) -> bytes:
        return config_structure.export().encode("utf-8")


def verify_section(
    section: Section, path_prefix: str = ""
) -> tuple[list[str], list[str], list[AttrTuple]]:
    # prevents putting in any characters which break corosync.conf structure
    bad_section_name_list = []
    bad_attribute_name_list = []
    bad_attribute_value_list = []

    path = _prefix_path(path_prefix, section.name)
    # the root section has no name yet it is valid
    if section.name != "" and not _is_valid_name(section.name):
        bad_section_name_list.append(path)
    for name, value in section.get_attributes():
        if not _is_valid_name(name):
            bad_attribute_name_list.append(_prefix_path(path, name))
        if not _is_valid_value(value):
            bad_attribute_value_list.append((_prefix_path(path, name), value))
    for child_section in section.get_sections():
        bad_sections, bad_attr_names, bad_attr_values = verify_section(
            child_section, path
        )
        bad_section_name_list += bad_sections
        bad_attribute_name_list += bad_attr_names
        bad_attribute_value_list += bad_attr_values

    return (
        bad_section_name_list,
        bad_attribute_name_list,
        bad_attribute_value_list,
    )


def _prefix_path(prefix: str, path: str) -> str:
    return f"{prefix}.{path}" if prefix and path else path


def _is_valid_name(name: str) -> bool:
    return constants.OPTION_NAME_RE.fullmatch(name) is not None


def _is_valid_value(value: str) -> bool:
    # pylint: disable=superfluous-parens
    return not (set(value) & set("{}\n\r"))


class CorosyncConfParserException(ParserErrorException):
    pass


class CircularParentshipException(CorosyncConfParserException):
    pass


class ParsingErrorException(CorosyncConfParserException):
    pass


class MissingClosingBraceException(ParsingErrorException):
    pass


class UnexpectedClosingBraceException(ParsingErrorException):
    pass


class MissingSectionNameBeforeOpeningBraceException(ParsingErrorException):
    pass


class ExtraCharactersAfterOpeningBraceException(ParsingErrorException):
    pass


class ExtraCharactersBeforeOrAfterClosingBraceException(ParsingErrorException):
    pass


class LineIsNotSectionNorKeyValueException(ParsingErrorException):
    pass



from typing import (
    Any,
    Optional,
)

from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.file import (
    FileMetadata,
    RawFileInterface,
)
from pcs.lib.file import (
    metadata,
    raw_file,
    toolbox,
)
from pcs.lib.interface.config import (
    FacadeInterface,
    ParserErrorException,
)


class FileInstance:
    """
    Provides a high-level easy access to config files
    """

    @classmethod
    def for_booth_config(
        cls,
        name: str,
        ghost_file: bool = False,
        ghost_data: Optional[bytes] = None,
    ) -> "FileInstance":
        """
        Factory for booth config file

        name -- booth instance name
        ghost_file -- use GhostFile if True, RealFile otherwise
        ghost_data -- initial GhostFile data
        """
        return cls._for_booth(
            file_type_codes.BOOTH_CONFIG, name, ghost_file, ghost_data
        )

    @classmethod
    def for_booth_key(
        cls,
        name: str,
        ghost_file: bool = False,
        ghost_data: Optional[bytes] = None,
    ) -> "FileInstance":
        """
        Factory for booth key file

        name -- booth instance name
        ghost_file -- use GhostFile if True, RealFile otherwise
        ghost_data -- initial GhostFile data
        """
        return cls._for_booth(
            file_type_codes.BOOTH_KEY, name, ghost_file, ghost_data
        )

    @classmethod
    def _for_booth(
        cls,
        file_type_code: file_type_codes.FileTypeCode,
        name: str,
        ghost_file: bool,
        ghost_data: Optional[bytes],
    ) -> "FileInstance":
        return cls(
            _get_raw_file(
                metadata.for_file_type(file_type_code, name),
                ghost_file,
                ghost_data,
            ),
            toolbox.for_file_type(file_type_code),
        )

    @classmethod
    def for_known_hosts(cls) -> "FileInstance":
        """
        Factory for known-hosts file
        """
        return cls._for_common(file_type_codes.PCS_KNOWN_HOSTS)

    @classmethod
    def for_pacemaker_key(cls) -> "FileInstance":
        """
        Factory for pacemaker key file
        """
        return cls._for_common(file_type_codes.PACEMAKER_AUTHKEY)

    @classmethod
    def for_dr_config(cls) -> "FileInstance":
        """
        Factory for disaster-recovery config file
        """
        return cls._for_common(file_type_codes.PCS_DR_CONFIG)

    @classmethod
    def for_corosync_conf(cls) -> "FileInstance":
        """
        Factory for corosync config file
        """
        return cls._for_common(file_type_codes.COROSYNC_CONF)

    @classmethod
    def for_pcs_users_config(cls) -> "FileInstance":
        return cls._for_common(file_type_codes.PCS_USERS_CONF)

    @classmethod
    def for_pcs_settings_config(cls) -> "FileInstance":
        return cls._for_common(file_type_codes.PCS_SETTINGS_CONF)

    @classmethod
    def _for_common(
        cls,
        file_type_code: file_type_codes.FileTypeCode,
    ) -> "FileInstance":
        return cls(
            raw_file.RealFile(metadata.for_file_type(file_type_code)),
            toolbox.for_file_type(file_type_code),
        )

    def __init__(
        self,
        raw_file_interface: RawFileInterface,
        file_toolbox: toolbox.FileToolbox,
    ):
        """
        Factories should be used instead

        RawFileInterface raw_file_interface -- RealFile or GhostFile instance
        FileToolbox file_toolbox -- collection of file tools
        """
        self._raw_file = raw_file_interface
        self._toolbox = file_toolbox

    @property
    def raw_file(self) -> RawFileInterface:
        """
        Get the underlying RawFileInterface instance
        """
        return self._raw_file

    @property
    def toolbox(self) -> toolbox.FileToolbox:
        """
        Get the underlying FileToolbox instance
        """
        return self._toolbox

    def parser_exception_to_report_list(
        self,
        exception: ParserErrorException,
        force_code: Optional[reports.types.ForceCode] = None,
        is_forced_or_warning: bool = False,
    ) -> reports.ReportItemList:
        """
        Translate a RawFileError instance to a report

        exception -- an exception to be translated
        string force_code -- is it a forcible error? by which code?
        is_forced_or_warning -- return a warning if True, error otherwise
        """
        return self._toolbox.parser.exception_to_report_list(
            exception,
            self._raw_file.metadata.file_type_code,
            (
                None
                if isinstance(self._raw_file, raw_file.GhostFile)
                else self._raw_file.metadata.path
            ),
            force_code,
            is_forced_or_warning,
        )

    def read_to_facade(self) -> FacadeInterface:
        return self._toolbox.facade(self.read_to_structure())

    def read_to_structure(self) -> Any:
        return self._toolbox.parser.parse(self.read_raw())

    def read_raw(self) -> bytes:
        return self._raw_file.read()

    def raw_to_facade(self, raw_file_data: bytes) -> FacadeInterface:
        """
        Parse raw file data and return a corresponding facade

        raw_file_data -- data to be parsed
        """
        return self._toolbox.facade(self.raw_to_structure(raw_file_data))

    def raw_to_structure(self, raw_file_data: bytes) -> Any:
        """
        Parse raw file data and return a corresponding structure

        raw_file_data -- data to be parsed
        """
        return self._toolbox.parser.parse(raw_file_data)

    def facade_to_raw(self, facade: FacadeInterface) -> bytes:
        return self._toolbox.exporter.export(facade.config)

    def write_facade(
        self, facade: FacadeInterface, can_overwrite: bool = False
    ) -> None:
        self.write_structure(facade.config, can_overwrite=can_overwrite)

    def write_structure(
        self, structure: Any, can_overwrite: bool = False
    ) -> None:
        self.write_raw(
            self._toolbox.exporter.export(structure),
            can_overwrite=can_overwrite,
        )

    def write_raw(
        self, raw_file_data: bytes, can_overwrite: bool = False
    ) -> None:
        """
        Write raw data to a file

        raw_file_data -- data to be written
        can_overwrite -- raise if False and the file already exists
        """
        self._raw_file.write(raw_file_data, can_overwrite=can_overwrite)


def _get_raw_file(
    file_metadata: FileMetadata, is_ghost: bool, ghost_data: Optional[bytes]
) -> RawFileInterface:
    if is_ghost:
        return raw_file.GhostFile(file_metadata, file_data=ghost_data)
    return raw_file.RealFile(file_metadata)



from unittest import (
    TestCase,
    mock,
)

from pcs.cli import cluster_property
from pcs.cli.common.errors import CmdLineInputError
from pcs.common.reports import codes as report_codes

from pcs_test.tools.misc import dict_to_modifiers


class TestSetProperty(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.Mock(spec_set=["set_properties"])
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.set_property(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.cluster_property.set_properties.assert_not_called()

    def test_empty_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["", ""])
        self.assertEqual(cm.exception.message, "missing value of '' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_option_missing_value(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a=1", "b"])
        self.assertEqual(cm.exception.message, "missing value of 'b' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_value_missing_option_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["=1", "b=2"])
        self.assertEqual(cm.exception.message, "missing key in '=1' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_option_multiple_values(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a=1", "a=2", "b=2", "b="])
        self.assertEqual(
            cm.exception.message,
            "duplicate option 'a' with different values '1' and '2'",
        )
        self.cluster_property.set_properties.assert_not_called()

    def test_multiple_args(self):
        self._call_cmd(["a=1", "b=2", "c="])
        self.cluster_property.set_properties.assert_called_once_with(
            {"a": "1", "b": "2", "c": ""}, set()
        )

    def test_multiple_args_with_force(self):
        self._call_cmd(["a=1", "b=2", "c="], {"force": True})
        self.cluster_property.set_properties.assert_called_once_with(
            {"a": "1", "b": "2", "c": ""}, set([report_codes.FORCE])
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--all' is not supported in this command",
        )
        self.cluster_property.assert_not_called()


class TestUnsetProperty(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.Mock(spec_set=["set_properties"])
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.unset_property(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.cluster_property.set_properties.assert_not_called()

    def test_args(self):
        self._call_cmd(["a=1", "=b", ""])
        self.cluster_property.set_properties.assert_called_once_with(
            {"a=1": "", "=b": "", "": ""}, set()
        )

    def test_args_with_force(self):
        self._call_cmd(["a=1", "=b", ""], {"force": True})
        self.cluster_property.set_properties.assert_called_once_with(
            {"a=1": "", "=b": "", "": ""}, set([report_codes.FORCE])
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"defaults": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--defaults' is not supported in this command",
        )
        self.cluster_property.assert_not_called()

    def test_duplicate_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a", "a", "b", "b", "c"])
        self.assertEqual(
            cm.exception.message,
            "duplicate arguments: 'a', 'b'",
        )
        self.cluster_property.assert_not_called()



import inspect
from dataclasses import (
    fields,
    is_dataclass,
)
from typing import (
    Container,
    Iterable,
    get_args,
    get_origin,
)
from unittest import TestCase

from pcs.daemon.async_tasks.worker.command_mapping import COMMAND_MAP


def prohibited_types_used(_type, prohibited_types):
    if _type in prohibited_types:
        return True
    generic = get_origin(_type)
    if generic:
        if generic in prohibited_types:
            return True
        return any(
            prohibited_types_used(arg, prohibited_types)
            for arg in get_args(_type)
        )
    if is_dataclass(_type):
        return any(
            prohibited_types_used(field.type, prohibited_types)
            for field in fields(_type)
        )
    return False


def _get_generic(annotation):
    return getattr(annotation, "__origin__", None)


class DaciteTypingCompatibilityTest(TestCase):
    def test_all(self):
        prohibited_types = (Iterable, Container)
        prohibited_types_normalized = [
            get_origin(_type) for _type in prohibited_types
        ]
        for cmd_name, cmd in COMMAND_MAP.items():
            for param in list(inspect.signature(cmd.cmd).parameters.values())[
                1:
            ]:
                if param.annotation != inspect.Parameter.empty:
                    self.assertFalse(
                        prohibited_types_used(
                            param.annotation, prohibited_types_normalized
                        ),
                        f"Prohibited type used in command: {cmd_name}; argument: {param}; prohibited_types: {prohibited_types}",
                    )



import os
import ssl

from pcs.common.ssl import (
    check_cert_key,
    dump_cert,
    dump_key,
    generate_cert,
    generate_key,
)


def _open_ssl_file_to_rewrite(path):
    return os.fdopen(os.open(path, os.O_CREAT | os.O_WRONLY, 0o600), "wb")


def regenerate_cert_key(server_name, cert_path, key_path, key_length=None):
    key = generate_key(key_length) if key_length else generate_key()
    with _open_ssl_file_to_rewrite(cert_path) as cert_file:
        cert_file.write(dump_cert(generate_cert(key, server_name)))
    with _open_ssl_file_to_rewrite(key_path) as key_file:
        key_file.write(dump_key(key))


class CertKeyPair:
    def __init__(self, cert_location, key_location):
        self.__cert_location = cert_location
        self.__key_location = key_location

    @property
    def cert_location(self):
        return self.__cert_location

    @property
    def key_location(self):
        return self.__key_location

    def exists(self):
        return os.path.isfile(self.cert_location) and os.path.isfile(
            self.key_location
        )

    def check(self):
        return check_cert_key(self.cert_location, self.key_location)

    def regenerate(self, server_name, key_length=None):
        regenerate_cert_key(
            server_name, self.cert_location, self.key_location, key_length
        )


class SSLCertKeyException(Exception):
    pass


class PcsdSSL:
    def __init__(
        self, server_name, cert_location, key_location, ssl_options, ssl_ciphers
    ):
        self.__server_name = server_name
        self.__ssl_options = ssl_options
        self.__ssl_ciphers = ssl_ciphers
        self.__ck_pair = CertKeyPair(cert_location, key_location)

    def create_context(self) -> ssl.SSLContext:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_context.set_ciphers(self.__ssl_ciphers)
        ssl_context.options = self.__ssl_options
        ssl_context.load_cert_chain(
            self.__ck_pair.cert_location, self.__ck_pair.key_location
        )
        return ssl_context

    def guarantee_valid_certs(self):
        if not self.__ck_pair.exists():
            self.__ck_pair.regenerate(self.__server_name)
            return

        error_list = self.__ck_pair.check()
        if error_list:
            raise SSLCertKeyException(*error_list)



from time import time as now
from typing import Optional

from pcs.common.tools import get_unique_uuid


class Session:
    def __init__(
        self,
        sid: str,
        username: str,
    ) -> None:
        # Session id propageted via cookies.
        self.__sid = sid
        # Username given by login attempt. It does not matter if the
        # authentication succeeded.
        self.__username = username
        # The moment of the last access. The only muttable attribute.
        self.refresh()

    @property
    def username(self) -> str:
        self.refresh()
        return self.__username

    @property
    def sid(self) -> str:
        self.refresh()
        return self.__sid

    def refresh(self) -> None:
        """
        Set the time of last access to now.
        """
        self.__last_access = now()

    def was_unused_last(self, seconds: int) -> bool:
        return now() > self.__last_access + seconds


class Storage:
    def __init__(self, lifetime_seconds: int) -> None:
        self.__sessions: dict[str, Session] = {}
        self.__lifetime_seconds = lifetime_seconds

    def get(self, sid: str) -> Optional[Session]:
        self.drop_expired()
        session = self.__sessions.get(sid)
        if session is not None:
            session.refresh()
        return session

    def drop_expired(self) -> None:
        obsolete_sid_list = [
            sid
            for sid, session in self.__sessions.items()
            if session.was_unused_last(self.__lifetime_seconds)
        ]
        for sid in obsolete_sid_list:
            self.destroy(sid)

    def destroy(self, sid: str) -> None:
        if sid in self.__sessions:
            del self.__sessions[sid]

    def login(self, username: str) -> Session:
        self.drop_expired()
        sid = get_unique_uuid(tuple(self.__sessions.keys()))
        session = Session(sid, username)
        self.__sessions[sid] = session
        return session



from unittest import TestCase

from pcs.cli.common import middleware


class MiddlewareBuildTest(TestCase):
    def test_run_middleware_correctly_chained(self):
        log = []

        def command(lib, argv, modifiers):
            log.append("command: {0}, {1}, {2}".format(lib, argv, modifiers))

        def mdw1(_next, lib, argv, modifiers):
            log.append("mdw1 start: {0}, {1}, {2}".format(lib, argv, modifiers))
            _next(lib, argv, modifiers)
            log.append("mdw1 done")

        def mdw2(_next, lib, argv, modifiers):
            log.append("mdw2 start: {0}, {1}, {2}".format(lib, argv, modifiers))
            _next(lib, argv, modifiers)
            log.append("mdw2 done")

        run_with_middleware = middleware.build(mdw1, mdw2)
        run_with_middleware(command, "1", "2", "3")
        self.assertEqual(
            log,
            [
                "mdw1 start: 1, 2, 3",
                "mdw2 start: 1, 2, 3",
                "command: 1, 2, 3",
                "mdw2 done",
                "mdw1 done",
            ],
        )



import logging
import logging.handlers
import os
import sys

# pylint: disable=import-error
import pyagentx

import pcs.utils
from pcs.snmp import settings
from pcs.snmp.updaters.v1 import ClusterPcsV1Updater

logger = logging.getLogger("pcs.snmp")
logger.addHandler(logging.NullHandler())


def is_debug():
    debug = os.environ.get("PCS_SNMP_AGENT_DEBUG", "")
    return debug.lower() in ["true", "on", "1"]


def get_update_interval():
    interval = os.environ.get("PCS_SNMP_AGENT_UPDATE_INTERVAL")
    if not interval:
        return settings.DEFAULT_UPDATE_INTERVAL

    def _log_invalid_value(_value):
        logger.warning(
            "Invalid update interval value: '%s' is not >= 1.0", str(_value)
        )
        logger.debug(
            "Using default update interval: %s",
            str(settings.DEFAULT_UPDATE_INTERVAL),
        )

    try:
        interval = float(interval)
    except ValueError:
        _log_invalid_value(interval)
        return settings.DEFAULT_UPDATE_INTERVAL
    if interval <= 1.0:
        _log_invalid_value(interval)
        return settings.DEFAULT_UPDATE_INTERVAL
    return interval


def setup_logging(debug=False):
    level = logging.INFO
    if debug:
        level = logging.DEBUG
        # this is required to enable debug also in the ruby code
        # key '--debug' has to be added
        pcs.utils.pcs_options["--debug"] = debug
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler = logging.handlers.WatchedFileHandler(
        settings.LOG_FILE, encoding="utf8"
    )
    handler.setLevel(level)
    handler.setFormatter(formatter)
    for logger_name in ["pyagentx", "pcs"]:
        logger_instance = logging.getLogger(logger_name)
        logger_instance.setLevel(level)
        logger_instance.addHandler(handler)


class PcsAgent(pyagentx.Agent):
    def setup(self):
        update_interval = get_update_interval()
        logger.info("Update interval set to: %s", str(update_interval))
        self.register(
            settings.PCS_OID + ".1",
            ClusterPcsV1Updater,
            freq=update_interval,
        )


def main():
    setup_logging(is_debug())
    try:
        agent = PcsAgent()
        agent.start()
    # pylint: disable=broad-except
    except Exception as e:
        print("Unhandled exception: {0}".format(str(e)))
        agent.stop()
        sys.exit(1)
    except KeyboardInterrupt:
        agent.stop()



from typing import Optional

from pcs import settings
from pcs.common import reports
from pcs.common.file import RawFileError
from pcs.common.str_tools import join_multilines
from pcs.lib.booth.config_facade import ConfigFacade
from pcs.lib.booth.constants import AUTHFILE_FIX_OPTION
from pcs.lib.booth.env import BoothEnv
from pcs.lib.errors import LibraryError
from pcs.lib.file.raw_file import raw_file_error_report
from pcs.lib.interface.config import ParserErrorException


def get_daemon_status(runner, name=None):
    cmd = [settings.booth_binary, "status"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    # 7 means that there is no booth instance running
    if return_value not in [0, 7]:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.BoothDaemonStatusError(
                    join_multilines([stderr, stdout])
                )
            )
        )
    return stdout


def get_tickets_status(runner, name=None):
    cmd = [settings.booth_binary, "list"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    if return_value != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.BoothTicketStatusError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def get_peers_status(runner, name=None):
    cmd = [settings.booth_binary, "peers"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    if return_value != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.BoothPeersStatusError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def check_authfile_misconfiguration(
    env: BoothEnv, report_processor: reports.ReportProcessor
) -> Optional[reports.item.ReportItemMessage]:
    if (
        not settings.booth_enable_authfile_set_enabled
        and not settings.booth_enable_authfile_unset_enabled
    ):
        return None
    if not env.config.raw_file.exists():
        return None
    try:
        facade: ConfigFacade = env.config.read_to_facade()
        if (
            settings.booth_enable_authfile_set_enabled
            and facade.get_authfile()
            and (facade.get_option(AUTHFILE_FIX_OPTION) or "0").lower()
            not in ("1", "yes", "on")
        ):
            return reports.messages.BoothAuthfileNotUsed(env.instance_name)
        if (
            settings.booth_enable_authfile_unset_enabled
            and not settings.booth_enable_authfile_set_enabled
            and facade.get_option(AUTHFILE_FIX_OPTION) is not None
        ):
            return reports.messages.BoothUnsupportedOptionEnableAuthfile(
                env.instance_name
            )
    except RawFileError as e:
        report_processor.report(
            raw_file_error_report(e, is_forced_or_warning=True)
        )
    except ParserErrorException as e:
        report_processor.report_list(
            env.config.parser_exception_to_report_list(
                e, is_forced_or_warning=True
            )
        )
    return None



from typing import Optional

from pcs.common import reports as report
from pcs.common.fencing_topology import TARGET_TYPE_NODE
from pcs.common.types import StringCollection
from pcs.lib.cib import fencing_topology as cib_fencing_topology
from pcs.lib.cib.tools import (
    get_fencing_topology,
    get_resources,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.state import ClusterState


def add_level(
    lib_env: LibraryEnvironment,
    level,
    target_type,
    target_value,
    devices,
    force_device=False,
    force_node=False,
):
    """
    Validate and add a new fencing level

    LibraryEnvironment lib_env -- environment
    int|string level -- level (index) of the new fencing level
    constant target_type -- the new fencing level target value type
    mixed target_value -- the new fencing level target value
    Iterable devices -- list of stonith devices for the new fencing level
    bool force_device -- continue even if a stonith device does not exist
    bool force_node -- continue even if a node (target) does not exist
    """
    cib = lib_env.get_cib()
    cib_fencing_topology.add_level(
        lib_env.report_processor,
        get_fencing_topology(cib),
        get_resources(cib),
        level,
        target_type,
        target_value,
        devices,
        ClusterState(lib_env.get_cluster_state()).node_section.nodes,
        force_device,
        force_node,
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()
    lib_env.push_cib()


def get_config(lib_env: LibraryEnvironment):
    """
    Get fencing levels configuration.

    Return a list of levels where each level is a dict with keys: target_type,
    target_value. level and devices. Devices is a list of stonith device ids.

    LibraryEnvironment lib_env -- environment
    """
    cib = lib_env.get_cib()
    return cib_fencing_topology.export(get_fencing_topology(cib))


def remove_all_levels(lib_env: LibraryEnvironment):
    """
    Remove all fencing levels
    LibraryEnvironment lib_env -- environment
    """
    cib_fencing_topology.remove_all_levels(
        get_fencing_topology(lib_env.get_cib())
    )
    lib_env.push_cib()


def remove_levels_by_params(
    lib_env: LibraryEnvironment,
    level=None,
    # TODO create a special type, so that it cannot accept any string
    target_type: Optional[str] = None,
    target_value=None,
    devices: Optional[StringCollection] = None,
    # TODO remove, deprecated backward compatibility layer
    ignore_if_missing: bool = False,
    # TODO remove, deprecated backward compatibility layer
    target_may_be_a_device: bool = False,
):
    """
    Remove specified fencing level(s).

    LibraryEnvironment lib_env -- environment
    int|string level -- level (index) of the fencing level to remove
    target_type -- the removed fencing level target value type
    mixed target_value -- the removed fencing level target value
    devices -- list of stonith devices of the removed fencing level
    ignore_if_missing -- when True, do not report if level not found
    target_may_be_a_device -- enables backward compatibility mode for old CLI
    """
    topology_el = get_fencing_topology(lib_env.get_cib())
    report_list = cib_fencing_topology.remove_levels_by_params(
        topology_el,
        level,
        target_type,
        target_value,
        devices,
        ignore_if_missing,
        validate_device_ids=(not target_may_be_a_device),
    )

    if not target_may_be_a_device or target_type != TARGET_TYPE_NODE:
        if lib_env.report_processor.report_list(report_list).has_errors:
            raise LibraryError()
        lib_env.push_cib()
        return

    # TODO remove, deprecated backward compatibility mode
    # CLI command parameters are: level, node, stonith, stonith... Both the
    # node and the stonith list are optional. If the node is omitted and the
    # stonith list is present, there is no way to figure it out, since there is
    # no specification of what the parameter is. Hence the pre-lib code tried
    # both. First it assumed the first parameter is a node. If that fence level
    # didn't exist, it assumed the first parameter is a device. Since it was
    # only possible to specify node as a target back then, this is enabled only
    # in that case.
    # CLI has no way to figure out what the first parameter is. Therefore, the
    # lib must try both cases if asked to do so.
    if not report.has_errors(report_list):
        lib_env.report_processor.report_list(report_list)
        lib_env.push_cib()
        return

    level_not_found = False
    for report_item in report_list:
        if (
            report_item.message.code
            == report.codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST
        ):
            level_not_found = True
            break
    if not level_not_found:
        lib_env.report_processor.report_list(report_list)
        raise LibraryError()

    target_and_devices = [target_value]
    if devices:
        target_and_devices.extend(devices)
    report_list_second = cib_fencing_topology.remove_levels_by_params(
        topology_el,
        level,
        None,
        None,
        target_and_devices,
        ignore_if_missing,
        validate_device_ids=(not target_may_be_a_device),
    )
    if not report.has_errors(report_list_second):
        lib_env.report_processor.report_list(report_list_second)
        lib_env.push_cib()
        return

    lib_env.report_processor.report_list(report_list)
    lib_env.report_processor.report_list(report_list_second)
    raise LibraryError()


def verify(lib_env: LibraryEnvironment):
    """
    Check if all cluster nodes and stonith devices used in fencing levels exist

    LibraryEnvironment lib_env -- environment
    """
    cib = lib_env.get_cib()
    lib_env.report_processor.report_list(
        cib_fencing_topology.verify(
            get_fencing_topology(cib),
            get_resources(cib),
            ClusterState(lib_env.get_cluster_state()).node_section.nodes,
        )
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()



import base64
from collections import namedtuple
from typing import (
    Any,
    Dict,
)

from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.errors import LibraryError


def create_pcmk_remote_actions(action_list):
    return {
        "pacemaker_remote {0}".format(action): service_cmd_format(
            "pacemaker_remote", action
        )
        for action in action_list
    }


def pcmk_authkey_format(authkey_content):
    """
    Return a dict usable in the communication with a remote/put_file
    authkey_content is raw authkey content
    """
    return {
        "data": base64.b64encode(authkey_content).decode("utf-8"),
        "type": "pcmk_remote_authkey",
        "rewrite_existing": True,
    }


def corosync_authkey_format(authkey_content):
    """
    Return a dict usable in the communication with a remote/put_file
    authkey_content is raw authkey content
    """
    return {
        "data": base64.b64encode(authkey_content).decode("utf-8"),
        "type": "corosync_authkey",
        "rewrite_existing": True,
    }


def pcmk_authkey_file(authkey_content):
    return {"pacemaker_remote authkey": pcmk_authkey_format(authkey_content)}


def corosync_authkey_file(authkey_content):
    return {"corosync authkey": corosync_authkey_format(authkey_content)}


def corosync_conf_format(corosync_conf_content):
    return {
        "type": "corosync_conf",
        "data": corosync_conf_content,
    }


def corosync_conf_file(corosync_conf_content):
    return {"corosync.conf": corosync_conf_format(corosync_conf_content)}


def pcs_dr_config_format(dr_conf_content: bytes) -> Dict[str, Any]:
    return {
        "type": "pcs_disaster_recovery_conf",
        "data": base64.b64encode(dr_conf_content).decode("utf-8"),
        "rewrite_existing": True,
    }


def pcs_dr_config_file(dr_conf_content: bytes) -> Dict[str, Any]:
    return {"disaster-recovery config": pcs_dr_config_format(dr_conf_content)}


def pcs_settings_conf_format(content):
    return {
        "data": content,
        "type": "pcs_settings_conf",
        "rewrite_existing": True,
    }


def pcs_settings_conf_file(content):
    return {"pcs_settings.conf": pcs_settings_conf_format(content)}


def service_cmd_format(service, command):
    """
    Return a dict usable in the communication with a remote/run_action
    string service is name of requested service (eg. pacemaker_remote)
    string command specifies an action on service (eg. start)
    """
    return {
        "type": "service_command",
        "service": service,
        "command": command,
    }


class Result(namedtuple("Result", "code message")):
    """Wrapper over some call results"""


def unpack_items_from_response(main_response, main_key, node_label):
    """
    Check format of main_response and return main_response[main_key].
    dict main_response has on the key 'main_key' dict with item name as key and
        dict with result as value. E.g.
        {
            "files": {
                "file1": {"code": "success", "message": ""}
            }
        }
    string main_key is name of key under that is a dict with results
    string node_label is a node label for reporting an invalid format
    """
    is_in_expected_format = (
        isinstance(main_response, dict)
        and main_key in main_response
        and isinstance(main_response[main_key], dict)
    )

    if not is_in_expected_format:
        raise LibraryError(
            ReportItem.error(reports.messages.InvalidResponseFormat(node_label))
        )

    return main_response[main_key]


def response_items_to_result(response_items, expected_keys, node_label):
    """
    Check format of response_items and return dict where keys are transformed to
    Result. E.g.
    {"file1": {"code": "success", "message": ""}}
    ->
    {"file1": Result("success", "")}}

    dict response_items has item name as key and dict with result as value.
    list expected_keys contains expected keys in a dict main_response[main_key]
    string node_label is a node label for reporting an invalid format
    """
    if set(expected_keys) != set(response_items.keys()):
        raise LibraryError(
            ReportItem.error(reports.messages.InvalidResponseFormat(node_label))
        )

    for result in response_items.values():
        if (
            not isinstance(result, dict)
            or "code" not in result
            or "message" not in result
        ):
            raise LibraryError(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(node_label)
                )
            )

    return {
        file_key: Result(raw_result["code"], raw_result["message"])
        for file_key, raw_result in response_items.items()
    }


def response_to_result(main_response, main_key, expected_keys, node_label):
    """
    Validate response (from remote/put_file or remote/run_action) and transform
    results from dict to Result.

    dict main_response has on the key 'main_key' dict with item name as key and
        dict with result as value. E.g.
        {
            "files": {
                "file1": {"code": "success", "message": ""}
            }
        }
    string main_key is name of key under that is a dict with results
    list expected_keys contains expected keys in a dict main_response[main_key]
    string node_label is a node label for reporting an invalid format
    """
    return response_items_to_result(
        unpack_items_from_response(main_response, main_key, node_label),
        expected_keys,
        node_label,
    )


def get_format_result(code_message_map):
    def format_result(result):
        if result.code in code_message_map:
            return code_message_map[result.code]

        return result.message

    return format_result



# pylint: disable=too-many-lines
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.tools import xml_fromstring
from pcs.lib.commands.resource import move_autoclean
from pcs.lib.xml_tools import etree_to_str

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import (
    TmpFileCall,
    TmpFileMock,
)
from pcs_test.tools.misc import get_test_resource as rc


def _node_fixture(name, node_id):
    return f'<node id="{node_id}" uname="{name}"/>'


def _node_list_fixture(nodes):
    return "\n".join(
        _node_fixture(node_name, node_id)
        for node_id, node_name in enumerate(nodes)
    )


def _nodes_section_fixture(content):
    return f"""
    <nodes>
    {content}
    </nodes>
    """


def _rsc_primitive_fixture(res_id):
    return f'<primitive id="{res_id}"/>'


def _rsc_clone_fixture(clone_id, primitive=None, is_promotable=False):
    if not primitive:
        primitive = _rsc_primitive_fixture(f"{clone_id}-primitive")
    meta_attrs = ""
    if is_promotable:
        meta_attrs = '<nvpair name="promotable" value="true" />'
    return f"""
        <clone id="{clone_id}">
            <meta_attributes id="{clone_id}-meta_attributes">
                {meta_attrs}
            </meta_attributes>
            {primitive}
        </clone>
    """


_resource_primitive = _rsc_primitive_fixture("A")

_resource_promotable_clone = _rsc_clone_fixture(
    "clone-A", primitive=_resource_primitive, is_promotable=True
)


def _resources_tag(content):
    return f"""
    <resources>
{content}
    </resources>
"""


def _state_resource_fixture(resource_id, role, node=None):
    node_list = []
    if node:
        node_list = [(node, 1)]

    return _resources_tag(
        """
        <resource id="{resource_id}" role="{role}" nodes_running_on="{nodes_count}">
            {nodes}
        </resource>
        """.format(
            resource_id=resource_id,
            role=role,
            nodes_count=len(node_list),
            nodes="\n".join(
                f"""<node name="{node_name}" id="{node_id}" cached="false" />"""
                for node_name, node_id in node_list
            ),
        )
    )


def _simulation_transition_fixture(synapses=""):
    return f"""
<transition_graph cluster-delay="60s" stonith-timeout="60s" failed-stop-offset="INFINITY" failed-start-offset="INFINITY" batch-limit="0" transition_id="0">
  {synapses}
</transition_graph>
"""


def _simulation_synapses_fixture(resource_id, start_id=0):
    return f"""
  <synapse id="{start_id}">
    <action_set>
      <rsc_op id="{start_id*10}" operation="stop" operation_key="{resource_id}_stop_0" on_node="node1" on_node_uuid="1">
        <primitive id="{resource_id}" class="ocf" provider="heartbeat" type="Dummy"/>
        <attributes CRM_meta_name="stop" CRM_meta_timeout="20000" crm_feature_set="3.0.9"/>
      </rsc_op>
    </action_set>
    <inputs/>
  </synapse>
  <synapse id="{start_id+1}">
    <action_set>
      <rsc_op id="{start_id*10+1}" operation="start" operation_key="{resource_id}_start_0" on_node="node2" on_node_uuid="2">
        <primitive id="{resource_id}" class="ocf" provider="heartbeat" type="Dummy"/>
        <attributes CRM_meta_name="start" CRM_meta_timeout="20000" crm_feature_set="3.0.9"/>
      </rsc_op>
    </action_set>
    <inputs>
      <trigger>
        <rsc_op id="{start_id*10}" operation="stop" operation_key="{resource_id}_stop_0" on_node="node1" on_node_uuid="1"/>
      </trigger>
    </inputs>
  </synapse>
"""


class MoveAutocleanCommonSetup(TestCase):
    def setUp(self):
        self.tmp_file_mock_obj = TmpFileMock(
            file_content_checker=assert_xml_equal,
        )
        self.addCleanup(self.tmp_file_mock_obj.assert_all_done)
        tmp_file_patcher = mock.patch("pcs.lib.tools.get_tmp_file")
        self.addCleanup(tmp_file_patcher.stop)
        tmp_file_mock = tmp_file_patcher.start()
        tmp_file_mock.side_effect = (
            self.tmp_file_mock_obj.get_mock_side_effect()
        )
        self.env_assist, self.config = get_env_tools(self)


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class MoveAutocleanSuccess(MoveAutocleanCommonSetup):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        super().setUp()

        self.orig_node = "node1"
        self.new_node = "node2"
        config_load_cib_name = "load_cib"
        self.config.runner.cib.load(
            name=config_load_cib_name,
            resources=_resources_tag(
                _resource_primitive + _resource_promotable_clone
            ),
            nodes=_nodes_section_fixture(
                _node_list_fixture([self.orig_node, self.new_node])
            ),
        )
        self.orig_cib = etree_to_str(
            xml_fromstring(self.config.calls.get(config_load_cib_name).stdout)
        )
        self.cib_with_constraint = '<updated_cib with_constraint="True"/>'
        self.cib_without_constraint = (
            '<cib with_constraint="False" updated="True"/>'
        )
        self.cib_simulate_constraint = (
            '<cib simulate="True" with_constraint="True"/>'
        )
        self.cib_rsc_move_tmp_file_name = "file_to_call_crm_resource_move"
        self.cib_diff_add_constraint_orig_tmp_file_name = (
            "cib_diff_add_constraint_orig"
        )
        self.cib_diff_add_constraint_updated_tmp_file_name = (
            "cib_diff_add_constraint_updated"
        )
        self.cib_constraint_removed_by_unmove_file_name = (
            "cib_constraint_removed_by_unmove"
        )
        self.cib_diff_remove_constraint_orig_tmp_file_name = (
            "cib_diff_remove_constraint_orig"
        )
        self.cib_diff_remove_constraint_updated_tmp_file_name = (
            "cib_diff_remove_constraint_updated"
        )
        self.simulated_cib_add_constraint_tmp_file_name = (
            "pcmk_simulate_move_new_cib"
        )
        self.simulated_transitions_add_constraint_tmp_file_name = (
            "pcmk_simulate_move_transitions"
        )
        self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name = (
            "simulated_cib_remove_constraint"
        )
        self.cib_simulated_apply_diff_removing_constraint = (
            '<cib simulate="True" with_constraint="False" diff_applied="True"/>'
        )
        self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name = (
            "pcmk_simulate_input_unmove_cib"
        )
        self.simulated_cib_remove_constraint_tmp_file_name = (
            "pcmk_simulate_unmove_new_cib"
        )
        self.simulated_transitions_remove_constraint_tmp_file_name = (
            "pcmk_simulate_unmove_transitions"
        )
        self.cib_apply_diff_remove_constraint_after_push_tmp_file_name = (
            "apply_cib_diff_remove_constraint"
        )
        self.cib_remove_constraint_diff_applied = (
            '<cib with_constraint="False" diff_applied="True"/>'
        )
        self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name = (
            "pcmk_simulate_after_push_input_unmove_cib_after"
        )
        self.simulated_cib_remove_constraint_after_push_tmp_file_name = (
            "pcmk_simulate_after_push_unmove_new_cib"
        )
        self.simulated_transitions_remove_constraint_after_push_tmp_file_name = (
            "pcmk_simulate_after_push_unmove_transitions"
        )
        self.cib_diff_add_constraint = "diff_add_constraint"
        self.cib_diff_remove_constraint = "diff_remove_constraint"

    def get_tmp_files(self, transitions):
        return [
            TmpFileCall(
                self.cib_rsc_move_tmp_file_name,
                orig_content=self.orig_cib,
                new_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_diff_add_constraint_orig_tmp_file_name,
                orig_content=self.orig_cib,
            ),
            TmpFileCall(
                self.cib_diff_add_constraint_updated_tmp_file_name,
                orig_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_constraint_removed_by_unmove_file_name,
                orig_content=self.cib_with_constraint,
                new_content=self.cib_without_constraint,
            ),
            TmpFileCall(
                self.cib_diff_remove_constraint_orig_tmp_file_name,
                orig_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_diff_remove_constraint_updated_tmp_file_name,
                orig_content=self.cib_without_constraint,
            ),
            TmpFileCall(
                self.simulated_cib_add_constraint_tmp_file_name,
                new_content=self.cib_simulate_constraint,
            ),
            TmpFileCall(
                self.simulated_transitions_add_constraint_tmp_file_name,
                new_content=transitions,
            ),
            TmpFileCall(
                self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name,
                orig_content=self.cib_simulate_constraint,
                new_content=self.cib_simulated_apply_diff_removing_constraint,
            ),
            TmpFileCall(
                self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name,
                orig_content=self.cib_simulate_constraint,
            ),
            TmpFileCall(
                self.simulated_cib_remove_constraint_tmp_file_name,
                new_content='<cib simulate="True" with_constraint="False"/>',
            ),
            TmpFileCall(
                self.simulated_transitions_remove_constraint_tmp_file_name,
                new_content=transitions,
            ),
            TmpFileCall(
                self.cib_apply_diff_remove_constraint_after_push_tmp_file_name,
                orig_content=self.cib_with_constraint,
                new_content=self.cib_remove_constraint_diff_applied,
            ),
            TmpFileCall(
                self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name,
                orig_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.simulated_cib_remove_constraint_after_push_tmp_file_name,
                new_content='<cib simulate="True" after_push="True" with_constraint="False"/>',
            ),
            TmpFileCall(
                self.simulated_transitions_remove_constraint_after_push_tmp_file_name,
                new_content=transitions,
            ),
        ]

    def set_up_testing_env(
        self,
        resource_id,
        status_before,
        status_after,
        with_node=False,
        is_promotable=False,
    ):
        self.config.runner.pcmk.load_state(
            resources=status_before,
            name="runner.pcmk.load_state.start",
        )
        self.config.runner.pcmk.resource_move(
            resource=resource_id,
            master=is_promotable,
            node=self.new_node if with_node else None,
            env=dict(CIB_file=self.cib_rsc_move_tmp_file_name),
        )
        self.config.runner.cib.diff(
            self.cib_diff_add_constraint_orig_tmp_file_name,
            self.cib_diff_add_constraint_updated_tmp_file_name,
            stdout=self.cib_diff_add_constraint,
            name="runner.cib.diff.add_constraint",
        )
        self.config.runner.pcmk.resource_clear(
            resource=resource_id,
            master=is_promotable,
            node=self.new_node if with_node else None,
            env=dict(CIB_file=self.cib_constraint_removed_by_unmove_file_name),
        )
        self.config.runner.cib.diff(
            self.cib_diff_remove_constraint_orig_tmp_file_name,
            self.cib_diff_remove_constraint_updated_tmp_file_name,
            stdout="diff_remove_constraint",
            name="runner.cib.diff.remove_constraint",
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_add_constraint_tmp_file_name,
            self.simulated_transitions_add_constraint_tmp_file_name,
            cib_xml=self.cib_with_constraint,
            name="pcmk.simulate.rsc.move",
        )
        self.config.runner.pcmk.load_state(
            resources=status_after,
            name="runner.pcmk.load_state.mid_simulation",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name
            ),
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.simulation.remove_constraint",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name
            ),
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_remove_constraint_tmp_file_name,
            self.simulated_transitions_remove_constraint_tmp_file_name,
            cib_xml=self.cib_simulated_apply_diff_removing_constraint,
            env=dict(
                CIB_file=self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name
            ),
            name="pcmk.simulate.rsc.unmove.on_simulated",
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_add_constraint,
            name="pcmk.push_cib_diff.live_cluster.add_constraint",
        )
        self.config.runner.pcmk.wait(
            timeout=0, name="runner.pcmk.wait.after_move"
        )
        self.config.runner.cib.load_content(
            self.cib_with_constraint,
            name="load_cib_after_move",
        )
        self.config.runner.pcmk.load_state(
            resources=status_after,
            name="runner.pcmk.load_state.after_push",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name
            ),
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.simulation.remove_constraint_after_move",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name
            ),
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_remove_constraint_after_push_tmp_file_name,
            self.simulated_transitions_remove_constraint_after_push_tmp_file_name,
            cib_xml=self.cib_remove_constraint_diff_applied,
            env=dict(
                CIB_file=self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name
            ),
            name="pcmk.simulate.rsc.unmove.after_push",
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.remove_constraint_after_move",
        )
        self.config.runner.pcmk.wait(
            timeout=0, name="runner.pcmk.wait.after_cleanup"
        )
        self.config.runner.pcmk.load_state(
            resources=status_after,
            name="runner.pcmk.load_state.final",
        )

    def get_reports(self, resource_id):
        return [
            fixture.info(
                reports.codes.RESOURCE_MOVE_CONSTRAINT_CREATED,
                resource_id=resource_id,
            ),
            fixture.info(
                reports.codes.WAIT_FOR_IDLE_STARTED,
                timeout=0,
            ),
            fixture.info(
                reports.codes.RESOURCE_MOVE_CONSTRAINT_REMOVED,
                resource_id=resource_id,
            ),
            fixture.info(
                reports.codes.WAIT_FOR_IDLE_STARTED,
                timeout=0,
            ),
            fixture.info(
                reports.codes.RESOURCE_RUNNING_ON_NODES,
                resource_id=resource_id,
                roles_with_nodes=dict(Started=[self.new_node]),
            ),
        ]

    def test_with_node(self):
        resource_id = "A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(f"other_than_{resource_id}")
                )
            )
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Started", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Started", self.new_node
            ),
            with_node=True,
        )
        move_autoclean(
            self.env_assist.get_env(), resource_id, node=self.new_node
        )
        self.env_assist.assert_reports(self.get_reports(resource_id))

    def test_no_strict(self):
        resource_id = "A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(f"other_than_{resource_id}")
                )
            )
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Started", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Started", self.new_node
            ),
        )
        move_autoclean(self.env_assist.get_env(), resource_id)
        self.env_assist.assert_reports(self.get_reports(resource_id))

    def test_strict(self):
        resource_id = "A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(_simulation_transition_fixture())
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Started", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Started", self.new_node
            ),
        )
        move_autoclean(self.env_assist.get_env(), resource_id, strict=True)
        self.env_assist.assert_reports(self.get_reports(resource_id))

    def test_master(self):
        resource_id = "clone-A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(_simulation_transition_fixture())
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Promoted", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Promoted", self.new_node
            ),
            is_promotable=True,
        )
        move_autoclean(self.env_assist.get_env(), resource_id, master=True)
        self.env_assist.assert_reports(
            self.get_reports(resource_id)[:-1]
            + [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    resource_id=resource_id,
                    roles_with_nodes=dict(Promoted=[self.new_node]),
                ),
            ]
        )

    def test_master_with_node(self):
        resource_id = "clone-A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(f"other_than_{resource_id}")
                )
            )
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Promoted", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Promoted", self.new_node
            ),
            with_node=True,
            is_promotable=True,
        )
        move_autoclean(
            self.env_assist.get_env(),
            resource_id,
            master=True,
            node=self.new_node,
        )
        self.env_assist.assert_reports(
            self.get_reports(resource_id)[:-1]
            + [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    resource_id=resource_id,
                    roles_with_nodes=dict(Promoted=[self.new_node]),
                ),
            ]
        )

    def test_master_with_node_strict(self):
        resource_id = "clone-A"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files(_simulation_transition_fixture())
        )
        self.set_up_testing_env(
            resource_id,
            status_before=_state_resource_fixture(
                resource_id, "Promoted", self.orig_node
            ),
            status_after=_state_resource_fixture(
                resource_id, "Promoted", self.new_node
            ),
            with_node=True,
            is_promotable=True,
        )
        move_autoclean(
            self.env_assist.get_env(),
            resource_id,
            master=True,
            node=self.new_node,
            strict=True,
        )
        self.env_assist.assert_reports(
            self.get_reports(resource_id)[:-1]
            + [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    resource_id=resource_id,
                    roles_with_nodes=dict(Promoted=[self.new_node]),
                ),
            ]
        )


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class MoveAutocleanValidations(MoveAutocleanCommonSetup):
    def test_cib_not_live(self):
        self.config.env.set_cib_data("<cib />")
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), "resource_id"),
            [
                fixture.error(
                    reports.codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=[file_type_codes.CIB],
                )
            ],
            expected_in_processor=False,
        )

    def test_resource_not_found(self):
        resource_id = "resource_id"
        self.config.runner.cib.load()
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id=resource_id,
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    context_type="resources",
                    context_id="",
                )
            ]
        )

    def test_resource_wrong_type_bundle(self):
        resource_id = "bundle_rsc"
        self.config.runner.cib.load(
            resources=_resources_tag(
                '<bundle id="{res_id}">{primitive}</bundle>'.format(
                    res_id=resource_id,
                    primitive=_rsc_primitive_fixture(
                        f"{resource_id}-primitive"
                    ),
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_BUNDLE,
                    resource_id=resource_id,
                )
            ]
        )

    def test_resource_wrong_type_clone(self):
        resource_id = "clone_rsc"
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_clone_fixture(resource_id))
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id=resource_id,
                )
            ]
        )

    def test_resource_wrong_type_inner_promotable_clone(self):
        resource_id = "A"
        promotable_id = f"{resource_id}-clone"
        self.config.runner.cib.load(
            resources=_resources_tag(
                _rsc_clone_fixture(
                    promotable_id,
                    primitive=_rsc_primitive_fixture(resource_id),
                    is_promotable=True,
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id=resource_id,
                    promotable_id=promotable_id,
                )
            ]
        )

    def test_resource_wrong_type_promoted_on_non_promotable_resource(self):
        resource_id = "A"
        promotable_id = f"{resource_id}-clone"
        self.config.runner.cib.load(
            resources=_resources_tag(
                _rsc_clone_fixture(
                    promotable_id,
                    primitive=_rsc_primitive_fixture(resource_id),
                    is_promotable=True,
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(), resource_id, master=True
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE,
                    resource_id=resource_id,
                    promotable_id=promotable_id,
                )
            ]
        )

    def test_stopped_resource_node_not_specified(self):
        resource_id = "A"
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(resource_id)),
        )
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(resource_id, "Stopped"),
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_NOT_RUNNING,
                    resource_id=resource_id,
                )
            ],
            expected_in_processor=False,
        )

    def test_stopped_resource_node_specified(self):
        resource_id = "A"
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(resource_id)),
        )
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(resource_id, "Stopped"),
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_NOT_RUNNING,
                    resource_id=resource_id,
                )
            ],
            expected_in_processor=False,
        )

    def test_node_not_found(self):
        resource_id = "A"
        node = "non_existing_node"
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(resource_id)),
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(), resource_id, node
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_NOT_FOUND,
                    node=node,
                    searched_types=[],
                )
            ],
        )

    def test_constraint_already_exist(self):
        resource_id = "A"
        config_load_cib_name = "load_cib"
        node = "node1"
        cib_with_constraint = '<cib with_constraint="True"/>'
        cib_without_constraint = '<cib with_constraint="False" updated="True"/>'
        cib_rsc_move_tmp_file_name = "cib_rsc_move_tmp_file"
        cib_diff_add_constraint_orig_tmp_file_name = (
            "cib_diff_add_constraint_orig"
        )
        cib_diff_add_constraint_updated_tmp_file_name = (
            "cib_diff_add_constraint_updated"
        )
        cib_diff_remove_constraint_orig_tmp_file_name = (
            "cib_diff_remove_constraint_orig"
        )
        cib_diff_remove_constraint_updated_tmp_file_name = (
            "cib_diff_remove_constraint_updated"
        )
        cib_constraint_removed_by_unmove_file_name = (
            "cib_constraint_removed_by_unmove"
        )
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(resource_id)),
            constraints=f"""
              <constraints>
                  <rsc_location id="prefer-{resource_id}" rsc="{resource_id}" role="Started" node="{node}" score="INFINITY"/>
              </constraints>
            """,
            nodes=_nodes_section_fixture(_node_list_fixture([node])),
            name=config_load_cib_name,
        )
        orig_cib = etree_to_str(
            xml_fromstring(self.config.calls.get(config_load_cib_name).stdout)
        )
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    cib_rsc_move_tmp_file_name,
                    orig_content=orig_cib,
                    new_content=cib_with_constraint,
                ),
                TmpFileCall(
                    cib_diff_add_constraint_orig_tmp_file_name,
                    orig_content=orig_cib,
                ),
                TmpFileCall(
                    cib_diff_add_constraint_updated_tmp_file_name,
                    orig_content=cib_with_constraint,
                ),
                TmpFileCall(
                    cib_constraint_removed_by_unmove_file_name,
                    orig_content=cib_with_constraint,
                    new_content=cib_without_constraint,
                ),
                TmpFileCall(
                    cib_diff_remove_constraint_orig_tmp_file_name,
                    orig_content=cib_with_constraint,
                ),
                TmpFileCall(
                    cib_diff_remove_constraint_updated_tmp_file_name,
                    orig_content=cib_without_constraint,
                ),
            ]
        )
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(resource_id, "Started", node),
        )
        self.config.runner.pcmk.resource_move(
            resource=resource_id,
            node=node,
            env=dict(CIB_file=cib_rsc_move_tmp_file_name),
        )
        self.config.runner.cib.diff(
            cib_diff_add_constraint_orig_tmp_file_name,
            cib_diff_add_constraint_updated_tmp_file_name,
            stdout="",
            name="runner.cib.diff.add_constraint",
        )
        self.config.runner.pcmk.resource_clear(
            resource=resource_id,
            node=node,
            env=dict(CIB_file=cib_constraint_removed_by_unmove_file_name),
        )
        self.config.runner.cib.diff(
            cib_diff_remove_constraint_orig_tmp_file_name,
            cib_diff_remove_constraint_updated_tmp_file_name,
            stdout="",
            name="runner.cib.diff.remove_constraint",
        )
        move_autoclean(self.env_assist.get_env(), resource_id, node=node)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.NO_ACTION_NECESSARY,
                ),
            ]
        )

    def test_unmanaged_resource(self):
        resource_id = "A"
        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(resource_id))
        )
        self.config.runner.pcmk.load_state(
            resources=_resources_tag(
                '<resource id="{resource_id}" managed="{managed}" />'.format(
                    resource_id=resource_id, managed="false"
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), resource_id),
            [
                fixture.error(
                    reports.codes.RESOURCE_IS_UNMANAGED,
                    resource_id=resource_id,
                )
            ],
            expected_in_processor=False,
        )


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class MoveAutocleanFailures(MoveAutocleanCommonSetup):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        super().setUp()

        self.resource_id = "A"
        self.affected_resource = f"other_than_{self.resource_id}"
        self.config_load_cib_name = "load_cib"
        self.cib_diff_add_constraint = "diff_add_constraint"
        self.cib_diff_remove_constraint = "diff_remove_constraint"
        self.cib_with_constraint = '<cib with_constraint="True"/>'
        self.cib_without_constraint = (
            '<cib with_constraint="False" updated="True"/>'
        )
        self.cib_rsc_move_tmp_file_name = "cib_rsc_move_tmp_file"
        self.cib_diff_add_constraint_orig_tmp_file_name = (
            "cib_diff_add_constraint_orig"
        )
        self.cib_diff_add_constraint_updated_tmp_file_name = (
            "cib_diff_add_constraint_updated"
        )
        self.cib_constraint_removed_by_unmove_file_name = (
            "cib_constraint_removed_by_unmove"
        )
        self.cib_diff_remove_constraint_orig_tmp_file_name = (
            "cib_diff_remove_constraint_orig"
        )
        self.cib_diff_remove_constraint_updated_tmp_file_name = (
            "cib_diff_remove_constraint_updated"
        )
        self.simulated_cib_add_constraint_tmp_file_name = (
            "pcmk_simulate_move_new_cib"
        )
        self.simulated_transitions_add_constraint_tmp_file_name = (
            "pcmk_simulate_move_transitions"
        )
        self.cib_simulate_constraint = (
            '<cib simulate="True" with_constraint="True"/>'
        )
        self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name = (
            "simulated_cib_remove_constraint"
        )
        self.cib_simulated_apply_diff_removing_constraint = (
            '<cib simulate="True" with_constraint="False" diff_applied="True"/>'
        )
        self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name = (
            "pcmk_simulate_input_unmove_cib"
        )
        self.simulated_cib_remove_constraint_tmp_file_name = (
            "pcmk_simulate_unmove_new_cib"
        )
        self.simulated_transitions_remove_constraint_tmp_file_name = (
            "pcmk_simulate_unmove_transitions"
        )
        self.cib_apply_diff_remove_constraint_after_push_tmp_file_name = (
            "apply_cib_diff_remove_constraint"
        )
        self.cib_remove_constraint_diff_applied = (
            '<cib with_constraint="False" diff_applied="True"/>'
        )
        self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name = (
            "pcmk_simulate_after_push_input_unmove_cib_after"
        )
        self.simulated_cib_remove_constraint_after_push_tmp_file_name = (
            "pcmk_simulate_after_push_unmove_new_cib"
        )
        self.simulated_transitions_remove_constraint_after_push_tmp_file_name = (
            "pcmk_simulate_after_push_unmove_transitions"
        )

        self.config.runner.cib.load(
            resources=_resources_tag(_rsc_primitive_fixture(self.resource_id)),
            nodes=_nodes_section_fixture(
                _node_list_fixture(["node1", "node2"])
            ),
            name=self.config_load_cib_name,
        )
        self.orig_cib = etree_to_str(
            xml_fromstring(
                self.config.calls.get(self.config_load_cib_name).stdout
            )
        )

    def get_tmp_files_mocks(
        self,
        simulation_transitions_move,
        simulation_transitions_unmove=None,
        simulation_transitions_unmove_after_push=None,
    ):
        file_list = [
            TmpFileCall(
                self.cib_rsc_move_tmp_file_name,
                orig_content=self.orig_cib,
                new_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_diff_add_constraint_orig_tmp_file_name,
                orig_content=self.orig_cib,
            ),
            TmpFileCall(
                self.cib_diff_add_constraint_updated_tmp_file_name,
                orig_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_constraint_removed_by_unmove_file_name,
                orig_content=self.cib_with_constraint,
                new_content=self.cib_without_constraint,
            ),
            TmpFileCall(
                self.cib_diff_remove_constraint_orig_tmp_file_name,
                orig_content=self.cib_with_constraint,
            ),
            TmpFileCall(
                self.cib_diff_remove_constraint_updated_tmp_file_name,
                orig_content=self.cib_without_constraint,
            ),
            TmpFileCall(
                self.simulated_cib_add_constraint_tmp_file_name,
                new_content=self.cib_simulate_constraint,
            ),
            TmpFileCall(
                self.simulated_transitions_add_constraint_tmp_file_name,
                new_content=simulation_transitions_move,
            ),
        ]

        if not simulation_transitions_unmove:
            return file_list

        file_list.extend(
            [
                TmpFileCall(
                    self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name,
                    orig_content=self.cib_simulate_constraint,
                    new_content=self.cib_simulated_apply_diff_removing_constraint,
                ),
                TmpFileCall(
                    self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name,
                    orig_content=self.cib_simulate_constraint,
                ),
                TmpFileCall(
                    self.simulated_cib_remove_constraint_tmp_file_name,
                    new_content='<cib simulate="True" with_constraint="False"/>',
                ),
                TmpFileCall(
                    self.simulated_transitions_remove_constraint_tmp_file_name,
                    new_content=simulation_transitions_unmove,
                ),
            ]
        )

        if not simulation_transitions_unmove_after_push:
            return file_list

        file_list.extend(
            [
                TmpFileCall(
                    self.cib_apply_diff_remove_constraint_after_push_tmp_file_name,
                    orig_content=self.cib_with_constraint,
                    new_content=self.cib_remove_constraint_diff_applied,
                ),
                TmpFileCall(
                    self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name,
                    orig_content=self.cib_with_constraint,
                ),
                TmpFileCall(
                    self.simulated_cib_remove_constraint_after_push_tmp_file_name,
                    new_content='<cib simulate="True" after_push="True" with_constraint="False"/>',
                ),
                TmpFileCall(
                    self.simulated_transitions_remove_constraint_after_push_tmp_file_name,
                    new_content=simulation_transitions_unmove_after_push,
                ),
            ]
        )

        return file_list

    def set_up_testing_env(self, node=None, stage=9):
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", "node1"
            ),
            name="runner.pcmk.load_state.before",
        )
        self.config.runner.pcmk.resource_move(
            resource=self.resource_id,
            node=node,
            env=dict(CIB_file=self.cib_rsc_move_tmp_file_name),
        )
        self.config.runner.cib.diff(
            self.cib_diff_add_constraint_orig_tmp_file_name,
            self.cib_diff_add_constraint_updated_tmp_file_name,
            stdout=self.cib_diff_add_constraint,
            name="runner.cib.diff.add_constraint",
        )
        self.config.runner.pcmk.resource_clear(
            resource=self.resource_id,
            node=node,
            env=dict(CIB_file=self.cib_constraint_removed_by_unmove_file_name),
        )
        self.config.runner.cib.diff(
            self.cib_diff_remove_constraint_orig_tmp_file_name,
            self.cib_diff_remove_constraint_updated_tmp_file_name,
            stdout=self.cib_diff_remove_constraint,
            name="runner.cib.diff.remove_constraint",
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_add_constraint_tmp_file_name,
            self.simulated_transitions_add_constraint_tmp_file_name,
            cib_xml=self.cib_with_constraint,
            name="pcmk.simulate.rsc.move",
        )
        if stage <= 1:
            return
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", node if node else "node2"
            ),
            name="runner.pcmk.load_state.mid_simulation",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name
            ),
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.simulation.remove_constraint",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name
            ),
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_remove_constraint_tmp_file_name,
            self.simulated_transitions_remove_constraint_tmp_file_name,
            cib_xml=self.cib_simulated_apply_diff_removing_constraint,
            env=dict(
                CIB_file=self.pcmk_simulate_remove_constraint_orig_cib_tmp_file_name
            ),
            name="pcmk.simulate.rsc.unmove.on_simulated",
        )
        if stage <= 2:
            return
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_add_constraint,
            name="pcmk.push_cib_diff.live_cluster.add_constraint",
        )
        self.config.runner.pcmk.wait(
            timeout=0, name="runner.pcmk.wait.after_move"
        )
        self.config.runner.cib.load_content(
            self.cib_with_constraint,
            name="load_cib_after_move",
        )
        if stage <= 3:
            return
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", node if node else "node2"
            ),
            name="runner.pcmk.load_state.after_push",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name
            ),
        )
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.simulation.remove_constraint_after_move",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name
            ),
        )
        self.config.runner.pcmk.simulate_cib(
            self.simulated_cib_remove_constraint_after_push_tmp_file_name,
            self.simulated_transitions_remove_constraint_after_push_tmp_file_name,
            cib_xml=self.cib_remove_constraint_diff_applied,
            env=dict(
                CIB_file=self.pcmk_simulate_remove_constraint_after_push_orig_cib_tmp_file_name
            ),
            name="pcmk.simulate.rsc.unmove.after_push",
        )
        if stage <= 4:
            return
        self.config.runner.cib.push_diff(
            cib_diff=self.cib_diff_remove_constraint,
            name="pcmk.push_cib_diff.remove_constraint_after_move",
        )
        self.config.runner.pcmk.wait(
            timeout=0, name="runner.pcmk.wait.after_cleanup"
        )

    def get_reports(self, stage=None):
        return [
            fixture.info(
                reports.codes.RESOURCE_MOVE_CONSTRAINT_CREATED,
                resource_id=self.resource_id,
            ),
            fixture.info(
                reports.codes.WAIT_FOR_IDLE_STARTED,
                timeout=0,
            ),
            fixture.info(
                reports.codes.RESOURCE_MOVE_CONSTRAINT_REMOVED,
                resource_id=self.resource_id,
            ),
            fixture.info(
                reports.codes.WAIT_FOR_IDLE_STARTED,
                timeout=0,
            ),
        ][: {None: None, 4: -2, 3: 2, 2: 0, 1: 0}[stage]]

    def test_move_affects_other_resources_strict(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(
                        self.affected_resource, start_id=1
                    )
                    + _simulation_synapses_fixture(self.resource_id)
                ),
            )
        )
        self.set_up_testing_env(stage=1)
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(), self.resource_id, strict=True
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_AFFECTS_OTRHER_RESOURCES,
                    resource_id=self.resource_id,
                    affected_resources=[self.affected_resource],
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.get_reports(stage=1))

    def test_unmove_affects_the_resource(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(
                        self.affected_resource, start_id=1
                    )
                    + _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
            )
        )
        self.set_up_testing_env(stage=2)
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), self.resource_id),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE,
                    resource_id=self.resource_id,
                    others_affected=False,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.get_reports(stage=2))

    def test_unmove_affects_other_resources_strict(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(
                        self.affected_resource, start_id=1
                    )
                ),
            )
        )
        self.set_up_testing_env(stage=2)
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
                strict=True,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE,
                    resource_id=self.resource_id,
                    others_affected=True,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.get_reports(stage=2))

    def test_unmove_after_push_affects_the_resource(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(
                        self.affected_resource, start_id=1
                    )
                    + _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.affected_resource)
                ),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
            )
        )
        setup_stage = 4
        self.set_up_testing_env(stage=setup_stage)
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(self.env_assist.get_env(), self.resource_id),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE,
                    resource_id=self.resource_id,
                    others_affected=False,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.get_reports(stage=setup_stage))

    def test_unmove_after_push_affects_other_resources_strict(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.affected_resource)
                ),
            )
        )
        setup_stage = 4
        self.set_up_testing_env(stage=setup_stage)
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
                strict=True,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE,
                    resource_id=self.resource_id,
                    others_affected=True,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.get_reports(stage=setup_stage))

    def test_resource_not_running_after_move(self):
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.affected_resource)
                ),
            )
        )
        self.set_up_testing_env()
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(self.resource_id, "Stopped"),
            name="runner.pcmk.load_state.final",
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
            ),
        )
        self.env_assist.assert_reports(
            self.get_reports()
            + [
                fixture.error(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id=self.resource_id,
                )
            ]
        )

    def test_simulation_resource_not_moved(self):
        node = "node2"
        different_node = f"different-{node}"
        setup_stage = 1
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
            )
            + [
                TmpFileCall(
                    self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name,
                    orig_content=self.cib_simulate_constraint,
                ),
            ]
        )
        self.set_up_testing_env(node=node, stage=setup_stage)
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", different_node
            ),
            name="runner.pcmk.load_state.final",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
                node=node,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_NOT_AFFECTING_RESOURCE,
                    resource_id=self.resource_id,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            self.get_reports(stage=setup_stage)
            + [
                fixture.debug(
                    reports.codes.TMP_FILE_WRITE,
                    file_path=self.cib_apply_diff_remove_constraint_from_simulated_cib_tmp_file_name,
                    content=self.cib_simulate_constraint,
                ),
            ]
        )

    def test_after_push_resource_not_moved(self):
        node = "node2"
        different_node = f"different-{node}"
        setup_stage = 3
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(),
            )
            + [
                TmpFileCall(
                    self.cib_apply_diff_remove_constraint_after_push_tmp_file_name,
                    orig_content=self.cib_with_constraint,
                ),
            ]
        )
        self.set_up_testing_env(node=node, stage=setup_stage)
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", different_node
            ),
            name="runner.pcmk.load_state.final",
            env=dict(
                CIB_file=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name,
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
                node=node,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_MOVE_NOT_AFFECTING_RESOURCE,
                    resource_id=self.resource_id,
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            self.get_reports(stage=setup_stage)
            + [
                fixture.debug(
                    reports.codes.TMP_FILE_WRITE,
                    file_path=self.cib_apply_diff_remove_constraint_after_push_tmp_file_name,
                    content=self.cib_with_constraint,
                ),
            ]
        )

    def test_resource_running_on_a_different_node(self):
        node = "node2"
        different_node = f"different-{node}"
        self.tmp_file_mock_obj.set_calls(
            self.get_tmp_files_mocks(
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.resource_id)
                ),
                _simulation_transition_fixture(),
                _simulation_transition_fixture(
                    _simulation_synapses_fixture(self.affected_resource)
                ),
            )
        )
        self.set_up_testing_env(node=node)
        self.config.runner.pcmk.load_state(
            resources=_state_resource_fixture(
                self.resource_id, "Started", different_node
            ),
            name="runner.pcmk.load_state.final",
        )
        self.env_assist.assert_raise_library_error(
            lambda: move_autoclean(
                self.env_assist.get_env(),
                self.resource_id,
                node=node,
            ),
        )
        self.env_assist.assert_reports(
            self.get_reports()
            + [
                fixture.error(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    resource_id=self.resource_id,
                    roles_with_nodes=dict(Started=[different_node]),
                )
            ]
        )



import json
import logging
from http.client import responses
from typing import (
    Any,
    Dict,
    Optional,
    Type,
    cast,
)

from dacite import (
    DaciteError,
    MissingValueError,
    UnexpectedDataError,
)
from tornado.web import (
    HTTPError,
    MissingArgumentError,
)

from pcs.common.async_tasks.dto import (
    CommandDto,
    TaskIdentDto,
)
from pcs.common.interface.dto import (
    DTOTYPE,
    PayloadConversionError,
    from_dict,
    to_dict,
)
from pcs.daemon.app.auth import (
    NotAuthorizedException,
    TokenAuthProvider,
)
from pcs.daemon.async_tasks.scheduler import (
    Scheduler,
    TaskNotFoundError,
)
from pcs.daemon.async_tasks.types import Command
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from .common import (
    BaseHandler,
    RoutesType,
)


class APIError(HTTPError):
    def __init__(
        self,
        http_code: int = 500,
        http_error: Optional[str] = None,
        error_msg: Optional[str] = None,
    ) -> None:
        super().__init__(http_code, reason=http_error)
        self.error_msg = error_msg


class RequestBodyMissingError(APIError):
    def __init__(self) -> None:
        super().__init__(
            400,
            error_msg="Request body is missing, has wrong format or "
            "wrong/missing headers.",
        )


class _BaseApiV2Handler(BaseHandler):
    """
    Base handler for the REST API

    Defines all common functions used by handlers, message body preprocessing,
    and HTTP(S) settings.
    """

    scheduler: Scheduler
    json: Optional[Dict[str, Any]] = None
    logger: logging.Logger
    _auth_provider: TokenAuthProvider

    def initialize(
        self, scheduler: Scheduler, auth_provider: AuthProvider
    ) -> None:
        super().initialize()
        self._auth_provider = TokenAuthProvider(self, auth_provider)
        self.scheduler = scheduler
        # TODO: Turn into a constant
        self.logger = logging.getLogger("pcs.daemon.scheduler")

    def prepare(self) -> None:
        """JSON preprocessing"""
        self.add_header("Content-Type", "application/json")
        if (
            "Content-Type" in self.request.headers
            and self.request.headers["Content-Type"] == "application/json"
        ):
            try:
                self.json = json.loads(self.request.body)
            except json.JSONDecodeError as exc:
                raise APIError(
                    http_code=400, error_msg="Malformed JSON data."
                ) from exc

    async def get_auth_user(self) -> AuthUser:
        try:
            return await self._auth_provider.auth_by_token()
        except NotAuthorizedException as e:
            raise APIError(http_code=401) from e

    @staticmethod
    def _from_dict_exc_handled(
        convert_to: Type[DTOTYPE], dictionary: Dict[str, Any]
    ) -> DTOTYPE:
        """
        Dacite conversion to DTO from JSON with handled exceptions
        :param convert_to: DTO type to return and validate against
        :return: DTO if JSON follows its structure, sends error response
            and connection ends otherwise
        """
        try:
            return from_dict(convert_to, dictionary, strict=True)
        except MissingValueError as exc:
            raise APIError(
                http_code=400,
                error_msg=f'Required key "{exc.field_path}" is missing in '
                f"request body.",
            ) from exc
        except UnexpectedDataError as exc:
            raise APIError(
                http_code=400,
                error_msg=f"Request body contains unexpected keys: "
                f"{', '.join(exc.keys)}.",
            ) from exc
        except (DaciteError, PayloadConversionError) as exc:
            raise APIError(
                http_code=400, error_msg="Malformed request body."
            ) from exc

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """
        JSON error responder for all API handlers

        This function provides unified error response for the whole API. This
        function is called when tornado encounters any Exception while this
        handler is being used. No need to call set_status in this method, it is
        already set by tornado.
        :param status_code: HTTP status code
        """
        response = {
            "http_code": status_code,
            "http_error": responses.get(status_code, "Unknown"),
            "error_message": None,
        }
        if "exc_info" in kwargs:
            _, exc, _ = kwargs["exc_info"]
            if isinstance(exc, HTTPError) and exc.reason:
                # Rewrite http reason autoconverted from http status code
                response["http_error"] = exc.reason
            if isinstance(exc, APIError):
                response["error_message"] = exc.error_msg

        self.finish(json.dumps(response))


class NewTaskHandler(_BaseApiV2Handler):
    """Create a new task from command"""

    async def post(self) -> None:
        auth_user = await self.get_auth_user()
        if self.json is None:
            raise RequestBodyMissingError()

        command_dto = self._from_dict_exc_handled(CommandDto, self.json)
        task_ident = self.scheduler.new_task(Command(command_dto), auth_user)
        self.write(json.dumps(to_dict(TaskIdentDto(task_ident))))


class RunTaskHandler(_BaseApiV2Handler):
    """Run command synchronously"""

    async def post(self) -> None:
        auth_user = await self.get_auth_user()
        if self.json is None:
            raise RequestBodyMissingError()

        command_dto = self._from_dict_exc_handled(CommandDto, self.json)
        task_ident = self.scheduler.new_task(Command(command_dto), auth_user)
        try:
            self.write(
                json.dumps(
                    to_dict(
                        await self.scheduler.wait_for_task(
                            task_ident, auth_user
                        )
                    )
                )
            )
        except TaskNotFoundError as exc:
            raise APIError(http_code=500) from exc


class TaskInfoHandler(_BaseApiV2Handler):
    """Get task status"""

    async def get(self) -> None:
        auth_user = await self.get_auth_user()
        try:
            task_ident = self.get_query_argument("task_ident")
            self.write(
                json.dumps(
                    to_dict(
                        self.scheduler.get_task(
                            cast(str, task_ident), auth_user
                        )
                    )
                )
            )
        except MissingArgumentError as exc:
            raise APIError(
                http_code=400,
                error_msg=f'URL argument "{exc.arg_name}" is missing.',
            ) from exc
        except TaskNotFoundError as exc:
            raise APIError(
                http_code=404,
                error_msg="Task with this identifier does not exist.",
            ) from exc


class KillTaskHandler(_BaseApiV2Handler):
    """Stop execution of a task"""

    async def post(self) -> None:
        auth_user = await self.get_auth_user()
        if self.json is None:
            raise RequestBodyMissingError()

        task_ident_dto = self._from_dict_exc_handled(TaskIdentDto, self.json)
        try:
            self.scheduler.kill_task(task_ident_dto.task_ident, auth_user)
        except TaskNotFoundError as exc:
            raise APIError(
                http_code=404,
                error_msg="Task with this identifier does not exist.",
            ) from exc
        self.set_status(200)
        self.finish()


def get_routes(
    scheduler: Scheduler,
    auth_provider: AuthProvider,
) -> RoutesType:
    """
    Returns mapping of URL routes to functions and links API to the scheduler
    :param scheduler: Scheduler's instance
    :return: URL to handler mapping
    """
    params = dict(scheduler=scheduler, auth_provider=auth_provider)
    return [
        ("/api/v2/task/result", TaskInfoHandler, params),
        ("/api/v2/task/create", NewTaskHandler, params),
        ("/api/v2/task/kill", KillTaskHandler, params),
        ("/api/v2/task/run", RunTaskHandler, params),
    ]
